#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#               2012      Tobias Bolten
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from threading import *
import sys, os
from gamera import core, util, config, classify
from gamera.plugins import features as features_module
import gamera.knncore, gamera.gamera_xml
import array

from gamera.knncore import CITY_BLOCK
from gamera.knncore import EUCLIDEAN
from gamera.knncore import FAST_EUCLIDEAN

KNN_XML_FORMAT_VERSION = 1.0

_distance_type_to_name = {
    CITY_BLOCK: "CITY-BLOCK",
    EUCLIDEAN: "EUCLIDEAN",
    FAST_EUCLIDEAN: "FAST-EUCLIDEAN" }

_distance_type_to_number = {
    "CITY-BLOCK": CITY_BLOCK,
    "EUCLIDEAN": EUCLIDEAN,
    "FAST-EUCLIDEAN": FAST_EUCLIDEAN }

# The kNN classifier stores it settings in a simple xml file -
# this class uses the gamera_xml.LoadXML class to load that
# file. After the file is loaded, kNN.load_settings extracts
# the data from the class to set up kNN.
class _KnnLoadXML(gamera.gamera_xml.LoadXML):
   def __init__(self):
      gamera.gamera_xml.LoadXML.__init__(self)

   def _setup_handlers(self):
      self.feature_functions = []
      self.selections = { }
      self.weights = { }
      self.num_k = None
      self.distance_type = None
      self.add_start_element_handler('gamera-knn-settings', self._tag_start_knn_settings)
      self.add_start_element_handler('selections', self._tag_start_selections)
      self.add_end_element_handler('selections', self._tag_end_selections)
      self.add_start_element_handler('weights', self._tag_start_weights)
      self.add_end_element_handler('weights', self._tag_end_weights)

   def _remove_handlers(self):
      self.remove_start_element_handler('gamera-knn-settings')
      self.remove_start_element_handler('selections')
      self.remove_start_element_handler('weights')

   def _tag_start_knn_settings(self, a):
      version = self.try_type_convert(a, 'version', float, 'gamera-knn-settings')
      if version < KNN_XML_FORMAT_VERSION:
         raise gamera_xml.XMLError(
            "knn-settings XML file is an older version that can not be read by this version of Gamera.")
      self.num_k = self.try_type_convert(a, 'num-k', int, 'gamera-knn-settings')
      self.distance_type = \
        _distance_type_to_number[self.try_type_convert(a, 'distance-type',
                                                       str, 'gamera-knn-settings')]

   def _tag_start_selections(self, a):
      self.add_start_element_handler('selection', self._tag_start_selection)
      self.add_end_element_handler('selection', self._tag_end_selection)

   def _tag_end_selections(self):
      self.remove_start_element_handler('selection')
      self.remove_start_element_handler('selection')

   def _tag_start_selection(self, a):
      self._data = u''
      self._selection_name = str(a["name"])
      self._parser.CharacterDataHandler = self._add_selections

   def _tag_end_selection(self):
      self._parser.CharacterDataHandler = None
      self.selections[self._selection_name] = array.array('i')
      nums = str(self._data).split()
      tmp = array.array('i', [int(x) for x in nums])
      self.selections[self._selection_name] = tmp

   def _add_selections(self, data):
      self._data += data

   def _tag_start_weights(self, a):
      self.add_start_element_handler('weight', self._tag_start_weight)
      self.add_end_element_handler('weight', self._tag_end_weight)

   def _tag_end_weights(self):
      self.remove_start_element_handler('weight')
      self.remove_end_element_handler('weight')

   def _tag_start_weight(self, a):
      self._data = u''
      self._weight_name = str(a["name"])
      self._parser.CharacterDataHandler = self._add_weights

   def _tag_end_weight(self):
      self._parser.CharacterDataHandler = None
      self.weights[self._weight_name] = array.array('d')
      nums = str(self._data).split()
      tmp = array.array('d', [float(x) for x in nums])
      self.weights[self._weight_name] = tmp

   def _add_weights(self, data):
      self._data += data

class _kNNBase(gamera.knncore.kNN):
   """k-NN classifier that supports optimization using
   a Genetic Algorithm. This classifier supports all of
   the Gamera interactive/non-interactive classifier interface."""

   def __init__(self, num_features=1, num_k=1, normalize=False):
      """Constructor for knn object. Features is a list
      of feature names to use for classification. If features
      is none then the default settings will be loaded from a
      file specified in the user config file. If there is no
      settings file specified then all of the features will be
      used."""
      gamera.knncore.kNN.__init__(self)
      self.num_features = num_features
      self.num_k = num_k
      self.normalize = normalize

   def __del__(self):
      pass

   def distance_from_images(self, images, glyph, max=None):
      """**distance_from_images** (ImageList *glyphs*, Image *glyph*, Float *max* = ``None``)

Compute a list of distances between a list of glyphs and a single glyph. Distances
greater than *max* are not included in the output.  The return value is a list of
floating-point distances.
"""
      self.generate_features(glyph)
      self.generate_features_on_glyphs(images)
      if max is None:
         return self._distance_from_images(iter(images), glyph)
      else:
         return self._distance_from_images(iter(images), glyph, max)

   def distance_between_images(self, imagea, imageb):
      """**distance_between_images** (Image *imagea*, Image *imageb*)

Compute the distance between two images using the settings
for the kNN object (distance_type, features, weights, etc). This
can be used when more control over the distance computations are
needed than with any of the other methods that work on multiple
images at once."""
      self.generate_features(imagea)
      self.generate_features(imageb)
      return self._distance_between_images(imagea, imageb)

   def distance_matrix(self, images, normalize=True):
      """**distance_matrix** (ImageList *images*, Bool *normalize* = ``True``)

Create a symmetric FloatImage containing all of the
distances between the images in the list passed in. This is useful
because it allows you to find the distance between any two pairs
of images regardless of the order of the pairs.

*normalize*
  When true, the features are normalized before performing the distance
  calculations."""
      self.generate_features_on_glyphs(images)
      l = len(images)
      progress = util.ProgressFactory("Generating unique distances...", l)
      m = self._distance_matrix(images, progress.step, normalize)
      #m = self._distance_matrix(images)
      progress.kill()
      return m

   def unique_distances(self, images, normalize=True):
      """**unique_distances** (ImageList *images*, Bool *normalize* = ``True``)

Return a list of the unique pairs of images in the passed in list
and the distances between them. The return list is a list of tuples
of (distance, imagea, imageb) so that it easy to sort.

*normalize*
  When true, the features are normalized before performing the distance
  calculations."""
      self.generate_features_on_glyphs(images)
      l = len(images)
      progress = util.ProgressFactory("Generating unique distances...", l)
      dists = self._unique_distances(images, progress.step, normalize)
      #dists = self._unique_distances(images)
      progress.kill()
      return dists

   def evaluate(self):
      """Float **evaluate** ()

Evaluate the performance of the kNN classifier using
leave-one-out cross-validation. The return value is a
floating-point number between 0.0 (0% correct) and 1.0 (100%
correct).
"""
      self.instantiate_from_images(self.database, self.normalize)
      ans = self.leave_one_out()
      return float(ans[0]) / float(ans[1])

   def knndistance_statistics(self, k=0):
      """**knndistance_statistics** (Int *k* = 0)

Returns a list of average distances between each training sample and its *k*
nearest neighbors. So, when you have *n* training samples, *n* average
distance values are returned. This can be useful for distance rejection.

Each item in the returned list is a tuple (*d*, *classname*), where
*d* is the average kNN distance and *classname* is the class name of the
training sample. In most cases, the class name is of little interest,
but it could be useful if you need class conditional distance statistics.
Beware however, that the average distance is computed over neighbors
belonging to any class, not just the same class. If you need the latter,
you must create a new classifier from training samples belonging only
to the specific class.

When *k* is zero, the property ``num_k`` of the knn classifier is used.
"""
      self.instantiate_from_images(self.database, self.normalize)
      progress = util.ProgressFactory("Generating knndistance statistics...", len(self.database))
      stats = self._knndistance_statistics(k, progress.step)
      progress.kill()
      return stats

   def settings_dialog(self, parent):
      """Display a settings dialog for k-NN settings"""
      from gamera import args
      dlg = args.Args([args.Int('k', range=(0, 100), default=self.num_k),
                       args.Choice('Distance Function',
                                   ['City block', 'Euclidean', 'Fast Euclidean'],
                                   default = self.distance_type)
                       ], name="kNN settings")
      results = dlg.show(parent)
      if results is None:
         return
      self.num_k, self.distance_type = results

   def save_settings(self, filename):
      """**save_settings** (FileSave *filename*)

Save the kNN settings to the given filename. This settings file (which is XML)
includes k, distance type, the current selection and weighting. This file is
different from the one produced by serialize in that it contains only the settings 
and no data."""
      from util import word_wrap
      file = open(filename, "w")
      indent = 0
      word_wrap(file, '<?xml version="1.0" encoding="utf-8"?>', indent)
      word_wrap(file,
                '<gamera-knn-settings version="%s" num-k="%s" distance-type="%s">'
                % (KNN_XML_FORMAT_VERSION,
                   self.num_k,
                   _distance_type_to_name[self.distance_type]), indent)
      indent += 1
      if self.feature_functions != None:
         # selections
         word_wrap(file, '<selections>', indent)
         indent += 1
         feature_no = 0
         selections = self.get_selections()
         for name, function in self.feature_functions[0]:
            word_wrap(file, '<selection name="%s">' % name, indent)
            length = function.return_type.length
            word_wrap(file, [x for x in
                             selections[feature_no:feature_no+length]],
                             indent + 1)
            word_wrap(file, '</selection>', indent)
            feature_no += length
         indent -= 1
         word_wrap(file, '</selections>', indent)

         # weights
         word_wrap(file, '<weights>', indent)
         indent += 1
         feature_no = 0
         weights = self.get_weights()
         for name, function in self.feature_functions[0]:
            word_wrap(file, '<weight name="%s">' % name, indent)
            length = function.return_type.length
            word_wrap(file,
                      [x for x in
                       weights[feature_no:feature_no+length]],
                      indent + 1)
            word_wrap(file, '</weight>', indent)
            feature_no += length
         indent -= 1
         word_wrap(file, '</weights>', indent)
      indent -= 1
      word_wrap(file, '</gamera-knn-settings>', indent)
      file.close()

   def load_settings(self, filename):
      """**load_settings** (FileOpen *filename*)

Load the kNN settings from an XML file.  See save_settings_."""
      from gamera import core

      loader = _KnnLoadXML()
      loader.parse_filename(filename)
      self.num_k = loader.num_k
      self.distance_type = loader.distance_type
      functions = loader.weights.keys()
      functions.sort()
      self.change_feature_set(functions)

      # Create the selection array with the selection in the correct order
      try:
         selections = array.array('i')
         for x in self.feature_functions[0]:
            selections.extend(loader.selections[x[0]])
         self.set_selections(selections)
      except Exception:
         # if no selections are given use the default values (all features are
         # selected)
         pass

      # Create the weights array with the weights in the correct order
      weights = array.array('d')
      for x in self.feature_functions[0]:
         weights.extend(loader.weights[x[0]])
      self.set_weights(weights)

   def serialize(self, filename):
      """**serialize** (FileSave *filename*)

Saves the classifier-specific settings *and* data in an optimized and
classifer-specific format.  

.. note:: 
   It is good practice to retain the XML
   file, since it is portable across platforms and to future versions of
   Gamera.  The binary format is not guaranteed to be portable."""
      if self.features == 'all':
         gamera.knncore.kNN.serialize(self, filename,['all'])
      else:
         gamera.knncore.kNN.serialize(self, filename,self.features)

   def unserialize(self, filename):
      """**unserialize** (FileOpen *filename*)

Opens the classifier-specific settings *and* data from an optimized and
classifer-specific format."""
      features = gamera.knncore.kNN.unserialize(self, filename)
      if len(features) == 1 and features[0] == 'all':
         self.change_feature_set('all')
      else:
         self.change_feature_set(features)

   def generate_features(self, glyph):
      """**generate_features** (Image *glyph*)

Generates features for the given glyph.
"""
      glyph.generate_features(self.feature_functions)

   def __get_settings_by_features(self, function):
      result = {}
      values = function()

      feature_no = 0
      for name, feature_function in self.feature_functions[0]:
         length = feature_function.return_type.length
         result[name] = [x for x in values[feature_no:feature_no+length]]
         feature_no += length

      return result

   def get_selections_by_features(self):
      """**get_selections_by_features** ()

Get the selection vector elements.

This function returns a python dictionary: keys are the feature names, values
are lists of zeros and ones, where ones correspond to the selected components.
"""
      return self.__get_settings_by_features(self.get_selections)

   def get_selections_by_feature(self, feature_name):
      """**get_selections_by_feature** (String *feature_name*)

Convenience wrapper for *get_selections_by_features* function. This function
returns only the selection values list for the given feature name.
"""
      selections = self.__get_settings_by_features(self.get_selections)

      if feature_name not in selections.keys():
         raise RuntimeError("get_selections_by_feature: feature is not in the feature set")

      return selections[feature_name]

   def get_weights_by_features(self):
      """**get_weights_by_features** ()

Get the weighting vector elements.

This function returns a python dictionary: keys are the feature names, values
are lists of real values in [0,1], which give the weight of the respective
component.
"""
      return self.__get_settings_by_features(self.get_weights)

   def get_weights_by_feature(self, feature_name):
      """**get_weights_by_feature** (String *feature_name*)

Convenience wrapper for *get_weights_by_features* function. This function
returns only the weighting values list for the given feature name.
"""
      weights = self.__get_settings_by_features(self.get_weights)

      if feature_name not in weights.keys():
         raise RuntimeError("get_weights_by_feature: feature is not in the feature set")

      return weights[feature_name]

   def __set_settings_by_features(self, function, values, a):
      if len(values.keys()) != len(self.feature_functions[0]):
         raise RuntimeError("set_settings_by_features: feature number mismatch")

      for name, feature_function in self.feature_functions[0]:
         if name not in values.keys():
            raise RuntimeError("set_settings_by_features: feature is not in the feature set")

         if len(values[name]) != feature_function.return_type.length:
            raise RuntimeError("set_settings_by_features: dimension from feature mismatch")

         # create the settings array
         a.extend(values[name])

      # apply the settings to the classifier
      function(a)

   def set_selections_by_features(self, values):
      """**set_selections_by_features** (Dictionary *values*)

Set the selection vector elements by the corresponding feature name.

*values*
   Python dictionary with feature names as keys and lists as values, as
   described in get_selections_by_features.

The dictionary must contain an entry for every feature of the currently
active feature set, that has been set in the constructor of the classifier or
by *change_feature_set*. Example:

.. code:: Python

   classifier = knn.kNNNonInteractive("train.xml",
                                      ["aspect_ratio","moments"], 0)
   classifier.set_selections_by_features({"aspect_ratio":[1],
                                          "moments":[0, 1, 1, 1, 1, 1, 1, 1, 0]})
"""
      a = array.array('i')
      self.__set_settings_by_features(self.set_selections, values, a)

   def set_selections_by_feature(self, feature_name, values):
      """ **set_selections_by_feature** (String *feature_name*, List *values*)

Set the selection vector elements for one specific feature.

*feature_name*
   The feature name as string.

*values*
   Python list with the selection values for the given feature. Dimension of the
   list must match with the feature dimension.
"""
      selections = self.get_selections_by_features()

      if feature_name not in selections.keys():
         raise RuntimeError("set_selections_by_feature: feature is not in the feature set")

      if not isinstance(values, list):
         raise RuntimeError("set_selections_by_feature: values is not a list")

      selections[feature_name] = values
      self.set_selections_by_features(selections)

   def set_weights_by_features(self, values):
      """**set_weights_by_features** (Dictionary *values*)

Set the weighing vector elements by the corresponding feature name.
The dictionary must contain an entry for every feature of the currently
active feature set, that has been set in the constructor of the classifier or
by *change_feature_set*. Example:

.. code:: Python

   classifier = knn.kNNNonInteractive("train.xml",
                                      ["aspect_ratio","moments"], 0)
   classifier.set_weights_by_features({"aspect_ratio":[0.6],
                                       "moments":[0.1, 1.0, 0.3, 0.5, 1.0, 0.0, 1.0, 0.9, 0.0]})
"""
      a = array.array('d')
      self.__set_settings_by_features(self.set_weights, values, a)

   def set_weights_by_feature(self, feature_name, values):
      """ **set_weights_by_feature** (String *feature_name*, List *values*)

Set the weighting vector elements for one specific feature.

*feature_name*
   The feature name as string.

*values*
   Python list with the weighting values for the given feature. Dimension of the
   list must match with the feature dimension.
"""
      weights = self.get_weights_by_features()

      if feature_name not in weights.keys():
         raise RuntimeError("set_weights_by_feature: feature is not in the feature set")

      if not isinstance(values, list):
         raise RuntimeError("set_weights_by_feature: values is not a list")

      weights[feature_name] = values
      self.set_weights_by_features(weights)

class kNNInteractive(_kNNBase, classify.InteractiveClassifier):
   def __init__(self, database=[], features='all', perform_splits=1, num_k=1):
      """**kNNInteractive** (ImageList *database* = ``[]``, *features* = 'all', bool *perform_splits* = ``True``, int *num_k* = ``1``)

Creates a new kNN interactive classifier instance.

*database*
        Must be a list (or Python interable) containing glyphs to use
        as training data for the classifier.

        Any images in the list that were manually classified (have
	classification_state == MANUAL) will be used as training data
	for the classifier.  Any UNCLASSIFIED or AUTOMATICALLY
	classified images will be ignored.

	When initializing a noninteractive classifier, the database
	*must* be non-empty.

*features*
	A list of feature function names to use for classification.
	These feature names
	correspond to the `feature plugin methods`__.  To use all
	available feature functions, pass in ``'all'``.

.. __: plugins.html#features

*perform_splits*
	  If ``perform_splits`` is ``True``, glyphs trained with names
	  beginning with ``_split.`` are run through a given splitting
	  algorithm.  For instance, glyphs that need to be broken into
	  upper and lower halves for further classification of those
	  parts would be trained as ``_split.splity``.  When the
	  automatic classifier encounters glyphs that most closely
	  match those trained as ``_split``, it will perform the
	  splitting algorithm and then continue to recursively
	  classify its parts.

	  The `splitting algorithms`__ are documented in the plugin documentation.

.. __: plugins.html#segmentation

          New splitting algorithms can be created by `writing plugin`__ methods
          in the category ``Segmentation``.  

.. __: writing_plugins.html

      """
      self.features = features
      self.feature_functions = core.ImageBase.get_feature_functions(features)
      num_features = features_module.get_features_length(features)
      _kNNBase.__init__(self, num_features=num_features, num_k=num_k)
      classify.InteractiveClassifier.__init__(self, database, perform_splits)

   def __del__(self):
      _kNNBase.__del__(self)
      classify.InteractiveClassifier.__del__(self)

   def noninteractive_copy(self):
      """**noninteractive_copy** ()

Creates a non-interactive copy of the interactive classifier."""
      return kNNNonInteractive(
         list(self.get_glyphs()), self.features, self._perform_splits, num_k=self.num_k)

   def supports_optimization(self):
      """Flag indicating that this classifier supports optimization."""
      return False

   def change_feature_set(self, f):
      """**change_feature_set** (*features*)

Changes the set of features used in the classifier to the given list of feature names.

*features*
  These feature names correspond to the `feature plugin methods`__.
  To use all available feature functions, pass in ``'all'``.

.. __: plugins.html#features"""
      self.features = f
      self.feature_functions = core.ImageBase.get_feature_functions(f)
      self.num_features = features_module.get_features_length(f)
      if len(self.database):
         self.is_dirty = True
         self.generate_features_on_glyphs(self.database)

class kNNNonInteractive(_kNNBase, classify.NonInteractiveClassifier):
   def __init__(self, database=[], features='all', perform_splits=True, num_k=1, normalize=False):
      """**kNNNonInteractive** (ImageList *database* = ``[]``, *features* = ``'all'``,
bool *perform_splits* = ``True``, int *num_k* = ``1``, bool *normalize* = ``False``)

Creates a new kNN classifier instance.

*database*
        Can be in one of two forms:

           - When a list (or Python iterable) each element is a glyph
             to use as training data for the classifier.  (For
             non-interactive classifiers, this list must be
             non-empty).

           - For non-interactive classifiers, *database* may be a
             filename, in which case the classifier will be
             "unserialized" from the given file.

        Any images in the list that were manually classified (have
	classification_state == MANUAL) will be used as training data
	for the classifier.  Any UNCLASSIFIED or AUTOMATICALLY
	classified images will be ignored.

	When initializing a noninteractive classifier, the database
	*must* be non-empty.

*features*
	A list of feature function names to use for classification.
	These feature names
	correspond to the `feature plugin methods`__.  To use all
	available feature functions, pass in ``'all'``.

.. __: plugins.html#features

*perform_splits*
	  If ``perform_splits`` is ``True``, glyphs trained with names
	  beginning with ``_split.`` are run through a given splitting
	  algorithm.  For instance, glyphs that need to be broken into
	  upper and lower halves for further classification of those
	  parts would be trained as ``_split.splity``.  When the
	  automatic classifier encounters glyphs that most closely
	  match those trained as ``_split``, it will perform the
	  splitting algorithm and then continue to recursively
	  classify its parts.

	  The `splitting algorithms`__ are documented in the plugin documentation.

.. __: plugins.html#segmentation

          New splitting algorithms can be created by `writing plugin`__ methods
          in the category ``Segmentation``.  

.. __: writing_plugins.html

*normalize*
    Normalize the feature vectors: x' = (x - mean_x)/stdev_x

      """
      self.features = features
      self.feature_functions = core.ImageBase.get_feature_functions(features)
      num_features = features_module.get_features_length(features)
      _kNNBase.__init__(self, num_features=num_features, num_k=num_k, normalize=normalize)
      classify.NonInteractiveClassifier.__init__(self, database, perform_splits)

   def __del__(self):
      _kNNBase.__del__(self)
      classify.NonInteractiveClassifier.__del__(self)

   def change_feature_set(self, f):
      """**change_feature_set** (*features*)

Changes the set of features used in the classifier to the given list of feature names.

*features*
  These feature names correspond to the `feature plugin methods`__.
  To use all available feature functions, pass in ``'all'``.

.. __: plugins.html#features"""
      self.features = f
      self.feature_functions = core.ImageBase.get_feature_functions(f)
      self.num_features = features_module.get_features_length(f)
      if len(self.database):
         self.is_dirty = True
         self.generate_features_on_glyphs(self.database)
         self.instantiate_from_images(self.database, self.normalize)

   def set_normalization_state(self, flag):
      """**set_normalization_state** (bool *flag*)
Set whether normalization is used or not for classification.
"""
      self.instantiate_from_images(self.database, flag)
      self.normalize = flag

   def get_normalization_state(self):
      """**get_normalization_state** ()
Returns whether normalization is used or not for classification.
"""
      return self.normalize

def simple_feature_selector(glyphs):
   """simple_feature_selector does a brute-force search through all
   possible combinations of features and returns a sorted list of
   tuples (accuracy, features). WARNING: this function should take
   a long time to complete."""

   if len(glyphs) <= 1:
      raise RuntimeError("Lenght of list must be greater than 1")
   
   c = classify.kNNNonInteractive()

   # For efficiency we calculate all of the features and pass in the
   # indexes of the features vector that we want to use for the distance
   # calculation. This is more efficient than using the features weights
   # or recalculating the features.
   all_features = []
   feature_indexes = {}
   offset = 0
   for x in glyphs[0].get_feature_functions()[0]:
      all_features.append(x[0])
      feature_indexes[x[0]] = range(offset, offset + x[1].return_type.length)
      offset += x[1].return_type.length
   # First do the easy ones = single features and all features
   answers = []
   c.change_feature_set(all_features)
   c.set_glyphs(glyphs)
   ans = c.classifier.leave_one_out()
   # Because we are only interested in top score, we can stop the evaluation
   # after we have missed too many answers to possibly beat the top score.
   # Therefore, we store the number missed for the top score here and pass
   # it into leave_one_out.
   stop_threshold = ans[1] - ans[0]
   answer = (float(ans[0]) / float(ans[1]), all_features)
   for x in all_features:
      ans = c.classifier.leave_one_out(feature_indexes[x], stop_threshold)
      num_wrong = ans[1] - ans[0]
      print num_wrong, ans[1], ans[0]
      if num_wrong < stop_threshold:
         stop_threshold = num_wrong
         answer = (float(ans[0]) / float(ans[1]), x)
   # Now do the remaining combinations using the CombGen object for each
   # size subset of all of the features.
   for i in range(2, len(all_features) - 1):
      for x in CombGen(all_features, i):
         indexes = []
         for y in x:
            indexes.extend(feature_indexes[y])
         ans = c.classifier.leave_one_out(indexes, stop_threshold)
         num_wrong = ans[1] - ans[0]
         print num_wrong, ans[1], ans[0]
         if num_wrong < stop_threshold:
            stop_threshold = num_wrong
            answer = (float(ans[0]) / float(ans[1]), x)
   return answer


class CombGen:
   """Generate the k-combinations of a sequence. This is a iterator
   that generates the combinations on the fly. This code was adapted
   from a posting by Tim Peters"""
   def __init__(self, seq, k):
      n = self.n = len(seq)
      if not 1 <= k <= n:
         raise ValueError("k must be in 1.." + `n` + ": " + `k`)
      self.k = k
      self.seq = seq
      self.indices = range(k)
      # Trickery to make the first .next() call work.
      self.indices[-1] = self.indices[-1] - 1

   def __iter__(self):
      return self

   def next(self):
      n, k, indices = self.n, self.k, self.indices
      lasti, limit = k-1, n-1
      while lasti >= 0 and indices[lasti] == limit:
         lasti = lasti - 1
         limit = limit - 1
      if lasti < 0:
         raise StopIteration
      newroot = indices[lasti] + 1
      indices[lasti:] = range(newroot, newroot + k - lasti)
      # Build the result.
      result = []
      seq = self.seq
      for i in indices:
         result.append(seq[i])
      return result

def _get_id_stats(glyphs, k=None):
   import stats
   if len(glyphs) < 3:
      return (len(glyphs),1.0, 1.0, 1.0)
   if k is None:
      k = kNN()
   distances = k.unique_distances(glyphs)
   return (len(glyphs),stats.lmean(distances), stats.lstdev(distances), stats.lmedian(distances))

def get_glyphs_stats(glyphs):
   k = kNN()
   klasses = {}
   for x in glyphs:
      id = x.get_main_id()
      if not klasses.has_key(id):
         klasses[id] = []
      klasses[id].append(x)
   stats = {}
   for x in klasses.iteritems():
      stats[x[0]] = _get_id_stats(x[1], k)
   return stats

def comma_delim_stats(glyphs, filename):
   file = open(filename, "w")
   stats = get_glyphs_stats(glyphs)
   for x in stats.iteritems():
      file.write(x[0])
      file.write(',')
      file.write(str(x[1][0]))
      file.write(',')
      file.write(str(x[1][1]))
      file.write(',')
      file.write(str(x[1][2]))
      file.write(',')
      file.write(str(x[1][3]))
      file.write('\n')
   file.close()

def glyphs_by_category(glyphs):
   klasses = {}
   for x in glyphs:
      id = x.get_main_id()
      if not klasses.has_key(id):
         klasses[id] = []
      klasses[id].append(x)
   return klasses



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                         and Karl MacMillan
#               2013      Manuel Jeltsch
#               2012-2013 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx
from wx.lib import buttons
import array
import os.path
import string
from gamera import util, enums
from gamera.gui import gui_util
from gamera.core import RGBPixel
from gamera.args import DEFAULT_MAX_ARG_NUMBER, CNoneDefault
import sys

class ArgInvalidException(Exception):
   pass

class Args:
   def _create_controls(self, locals, parent):
      self.controls = []
      # Controls
      if util.is_sequence(self.list[0]):
         notebook = wx.Notebook(parent, -1)
         for page in self.list:
            panel = wx.Panel(notebook, -1)
            gs = self._create_page(locals, panel, page[1:])
            panel.SetSizer(gs)
            notebook.AddPage(panel, page[0])
         return notebook
      else:
         return self._create_page(locals, parent, self.list)

   def _create_page(self, locals, parent, page):
      sw = wx.ScrolledWindow(
         parent, style=wx.SIMPLE_BORDER,
         size=(-1, -1))
      gs = self._create_page_impl(locals, sw, page)
      sw.SetSizer(gs)
      if wx.VERSION < (2, 9):
          gs.SetVirtualSizeHints(sw)
      else:
          gs.FitInside(sw)
      return sw

   def _create_page_impl(self, locals, parent, page):
      gs = wx.FlexGridSizer(len(page), 2, 2, 2)
      for item in page:
         if item.name == None:
            item.name = "ERROR!  No name given!"
         gs.Add(wx.StaticText(parent, -1, item.name),
                0,
                (wx.TOP|wx.LEFT|wx.RIGHT|
                 wx.EXPAND|
                 wx.ALIGN_CENTER_VERTICAL|
                 wx.ALIGN_LEFT), 10)
         control = item.get_control(parent, locals)
         self.controls.append(control)
         gs.Add(control.control,
                0,
                (wx.TOP|wx.LEFT|wx.RIGHT|
                 wx.EXPAND|
                 wx.ALIGN_CENTER_VERTICAL|
                 wx.ALIGN_RIGHT), 10)
      gs.AddGrowableCol(1)
      return gs

   def _create_buttons(self):
      # Buttons
      buttons = wx.BoxSizer(wx.HORIZONTAL)
      ok = wx.Button(self.window, wx.ID_OK, "OK")
      ok.SetDefault()
      buttons.AddStretchSpacer(1)
      buttons.Add(ok, 0, wx.ALL, 5)
      buttons.Add(wx.Button(self.window,
                           wx.ID_CANCEL,
                           "Cancel"),
                  0,
                  wx.ALL,
                  5)
      self.buttons = buttons
      return buttons

   def _create_wizard_buttons(self):
      # Buttons
      buttons = wx.BoxSizer(wx.HORIZONTAL)
      buttons.AddStretchSpacer(1)
      buttons.Add(wx.Button(self.window,
                           wx.ID_CANCEL,
                           "< Back"),
                  0,
                  wx.ALL,
                  5)
      ok = wx.Button(self.window,
                    wx.ID_OK,
                    "Next >")
      ok.SetDefault()
      buttons.Add(ok,
                  0,
                  wx.ALL,
                  5)
      return buttons

   if wx.VERSION >= (2, 5):
      import wx.html
      def _create_help_display(self, docstring):
         try:
            docstring = util.dedent(docstring)
            html = gui_util.docstring_to_html(docstring)
            window = wx.html.HtmlWindow(self.window, -1, size=wx.Size(50, 100))
            if wx.VERSION >= (2, 5) and "gtk2" in wx.PlatformInfo:
               window.SetStandardFonts()
            window.SetPage(html)
            window.SetBackgroundColour(wx.Colour(255, 255, 232))
            if wx.VERSION < (2, 8):
               window.SetBestFittingSize(wx.Size(50, 150))
            else:
               window.SetInitialSize(wx.Size(50, 150))
            return window
         except Exception, e:
            print e
   else:
      def _create_help_display(self, docstring):
         docstring = util.dedent(docstring)
         style = (wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH2)
         window = wx.TextCtrl(self.window, -1, style=style, size=wx.Size(50, 100))
         window.SetValue(docstring)
         window.SetBackgroundColour(wx.Colour(255, 255, 232))
         return window

   # generates the dialog box
   def setup(self, parent, locals, docstring = "", function = None):
      if function is not None:
         name = function
      else:
         name = self.name
      self.window = wx.Dialog(parent, -1, name,
                              style=wx.CAPTION|wx.RESIZE_BORDER)
      self.window.SetAutoLayout(1)
      if self.wizard:
         bigbox = wx.BoxSizer(wx.HORIZONTAL)
         from gamera.gui import gamera_icons
         bmp = gamera_icons.getGameraWizardBitmap()
         bitmap = wx.StaticBitmap(self.window, -1, bmp)
         bigbox.Add(bitmap, 0, wx.ALIGN_TOP)
      self.box = wx.BoxSizer(wx.VERTICAL)
      self.border = wx.BoxSizer(wx.HORIZONTAL)
      self.window.SetSizer(self.border)
      self.gs = self._create_controls(locals, self.window)
      if self.wizard:
         buttons = self._create_wizard_buttons()
      else:
         buttons = self._create_buttons()
      # Put it all together
      if self.title != None:
         static_text = wx.StaticText(self.window, 0, self.title)
         font = wx.Font(12, wx.SWISS, wx.NORMAL, wx.BOLD, False, "Helvetica")
         static_text.SetFont(font)
         self.box.Prepend(static_text, 0,
                      wx.EXPAND|wx.BOTTOM, 20)
      self.box.Add(self.gs, 1,
                   wx.EXPAND|wx.ALIGN_RIGHT)
      self.box.Add(wx.Panel(self.window, -1, size=(20,20)), 0,
                   wx.ALIGN_RIGHT)
      if docstring:
         help = self._create_help_display(docstring)
         self.box.Add(help, 1, wx.EXPAND)
      self.box.Add(buttons, 0, wx.ALIGN_RIGHT|wx.EXPAND)
      if self.wizard:
         bigbox.Add(
            self.box, 1,
            wx.EXPAND|wx.ALL|wx.ALIGN_TOP,
            15)
         self.border.Add(
            bigbox, 1, wx.EXPAND|wx.ALL, 10)
      else:
         self.border.Add(
            self.box, 1, wx.EXPAND|wx.ALL, 15)
      self.border.Fit(self.window)
      self.border.Layout()
      size = self.window.GetSize()
      self.window.SetSize((max(450, size[0]), max(200, size[1])))
      min_width = self.border.GetMinSize().width
      #self.border.Layout()
      #self.border.Fit(self.window)
      min_height = self.border.GetMinSize().height
      display = wx.Display(0)
      client_area = display.GetClientArea()
      if min_width < 350:
         min_width = 350
      if min_height < 200:
         min_height = 200
      if min_height < client_area.height/100*98:
         if wx.VERSION < (2, 8):
            self.gs.SetBestFittingSize(wx.Size(0, 0))
            self.window.SetBestFittingSize(wx.Size(min_width,min_height))
            self.gs.SetWindowStyle(wx.BORDER_NONE)
         else:
            height = self.window.GetSize().height
            self.gs.SetInitialSize(wx.Size(0, 0))
            self.window.SetInitialSize(wx.Size(min_width,height))
            self.gs.SetWindowStyle(wx.BORDER_NONE)
      else:
         if wx.VERSION < (2, 8):
            self.gs.SetBestFittingSize(wx.Size(0, 0))
            self.window.SetBestFittingSize(wx.Size(min_width,client_area.height/2))
            self.gs.EnableScrolling(0, 1)
            self.gs.SetScrollRate(0, 20)
         else:
            self.gs.SetInitialSize(wx.Size(0, 0))
            self.window.SetInitialSize(wx.Size(min_width,client_area.height/2))
            self.gs.EnableScrolling(0, 1)
            self.gs.SetScrollRate(0, 20)
      self.border.Layout()
      if wx.VERSION < (2, 9):
          self.border.SetVirtualSizeHints(self.window)
      else:
          self.border.FitInside(self.window)
      #self.border.Fit(self.window)
      self.window.Centre()

   def get_args_string(self):
      results = [x.get_string() for x in self.controls]
      tuple = '(' + ', '.join(results) + ')'
      return tuple

   def get_args(self):
      return [control.get() for control in self.controls]

   def show(self, parent=None, locals={}, function=None, wizard=0, docstring=""):
      docstring = util.dedent(docstring)
      self.wizard = wizard
      if function != None:
         self.function = function
      self.setup(parent, locals, docstring, function)
      while 1:
         result = wx.Dialog.ShowModal(self.window)
         try:
            if result == wx.ID_CANCEL:
               return None
            elif self.function is None:
               if function is None:
                  return tuple(self.get_args())
               else:
                  return function + self.get_args_string()
            else:
               ret = self.function + self.get_args_string()
               self.window.Destroy()
               return ret 
         except ArgInvalidException, e:
            gui_util.message(str(e))
         #except Exception:
         #   throw
         else:
            break
      self.window.Destroy()

class _NumericValidator(wx.PyValidator):
   def __init__(self, name="Float entry box ", range=None):
      wx.PyValidator.__init__(self)
      self.rng = range
      self.name = name
      wx.EVT_CHAR(self, self.OnChar)

   def Clone(self):
      return self.__class__(self.name, self.rng)

   def show_error(self, s):
      dlg = wx.MessageDialog(
         self.GetWindow(), s,
         "Dialog Error",
         wx.OK | wx.ICON_ERROR)
      dlg.ShowModal()
      dlg.Destroy()

   def Validate(self, win):
      tc = self.GetWindow()
      val = str(tc.GetValue())
      if val == "None":
          pass
      else:
          for x in val:
             if x not in self._digits:
                self.show_error(self.name + " must be numeric.")
                return False
          try:
             val = self._type(val)
          except Exception:
             self.show_error(self.caption + " is invalid.")
             return False
          if self.rng:
             if val < self.rng[0] or val > self.rng[1]:
                self.show_error(
                   self.name + " must be in the range " + str(self.rng) + ".")
                return False
      return True

   def OnChar(self, event):
      key = event.GetKeyCode()

      if (key < wx.WXK_SPACE or
          key == wx.WXK_DELETE or key > 255):
         event.Skip()
         return
      if chr(key) in self._digits:
         event.Skip()
         return
      if not wx.Validator_IsSilent():
         wx.Bell()

   def TransferToWindow(self):
      return True

   def TransferFromWindow(self):
      return True

class _IntValidator(_NumericValidator):
   _digits = string.digits + "-"
   _type = int

class Int:
   def get_control(self, parent, locals=None):
      self.control = wx.TextCtrl(parent, -1, str(self.default))
      self.control.SetValidator(_IntValidator(name = self.name, range = self.rng))
      return self

   def get(self):
      return int(self.control.GetValue())

   def get_string(self):
      return str(self.control.GetValue())

class _RealValidator(_NumericValidator):
   _digits = string.digits + "-."
   _type = float

class Real:
   def get_control(self, parent, locals=None):
      self.control = wx.TextCtrl(
         parent, -1, str(self.default),
         validator=_RealValidator(name=self.name, range=self.rng))
      return self

   def get(self):
      return float(self.control.GetValue())

   def get_string(self):
      return str(self.control.GetValue())

class String:
   def get_control(self, parent, locals=None):
      self.control = wx.TextCtrl(parent, -1, str(self.default))
      return self

   def get(self):
      return self.control.GetValue()

   def get_string(self):
      return "r'" + self.control.GetValue() + "'"

class Class:
   def determine_choices(self, locals):
      self.locals = locals
      if self.klass is None:
         choices = locals.keys()
      else:
         choices = []
         if self.list_of:
            for key, val in locals.items():
               try:
                  it = iter(val)
               except Exception:
                  pass
               else:
                  good = True
                  try:
                     for x in val:
                        if not isinstance(x, self.klass):
                           good = False
                           break
                  except Exception:
                     pass
                  else:
                     if good:
                        choices.append(key)
         else:
            for key, val in locals.items():
               if isinstance(val, self.klass):
                  choices.append(key)
      if isinstance(self.default, CNoneDefault):
         choices = ["None"] + choices
      return choices

   def get_control(self, parent, locals=None):
      if util.is_string_or_unicode(self.klass):
         self.klass = eval(self.klass)
      self.control = wx.Choice(
         parent, -1, choices = self.determine_choices(locals))
      #self.control.SetSelection(0)
      return self

   def get(self):
      try:
         return self.locals[self.control.GetStringSelection()]
      except Exception:
         if self.list_of:
            return []
         else:
            return None

   def get_string(self):
      try:
         return self.control.GetStringSelection()
      except Exception:
         if self.list_of:
            return []
         else:
            return None

class _Vector(Class):
   def get_control(self, parent, locals=None):
      if util.is_string_or_unicode(self.klass):
         self.klass = eval(self.klass)
      self.choices = self.determine_choices(locals)
      self.control = wx.ComboBox( 
        parent, -1, str(self.default), choices=self.choices, style=wx.CB_DROPDOWN)
      return self

   def get(self):
      value = self.control.GetValue()
      if value in self.choices:
         return self.locals[self.control.GetStringSelection()]
      else:
         try:
            x = eval(value)
         except SyntaxError:
            raise ArgInvalidException("Syntax error in '%s'.  Must be a %s" % (value, self.__class__.__name__))
         if not self.is_vector(x):
            raise ArgInvalidException("Argument '%s' must be a %s" % (value, self.__class__.__name__))
         return x

   def get_string(self):
      value = self.control.GetValue()
      if value in self.choices:
         return value
      else:
         try:
            x = eval(value)
         except SyntaxError:
            raise ArgInvalidException("Syntax error in '%s'.  Must be a %s" % (value, self.__class__.__name__))
         if not self.is_vector(x):
            raise ArgInvalidException("Argument '%s' must be a %s" % (value, self.__class__.__name__))
         return value

   def is_vector(self, val):
      try:
         it = iter(val)
      except Exception:
         return False
      else:
         good = True
         try:
            for x in val:
               if not isinstance(x, self.klass):
                  good = False
                  break
         except Exception:
            return False
         else:
            return good

   def determine_choices(self, locals):
      self.locals = locals
      choices = []
      for key, val in locals.items():
         if (isinstance(val, array.array) and
             val.typecode == self.typecode):
            choices.append(key)
         else:
            if self.is_vector(val):
               choices.append(key)
      return choices

class ImageType(Class):
   def determine_choices(self, locals):
      from gamera import core
      choices = []
      self.locals = locals
      if locals:
         if self.list_of:
            for key, val in locals.items():
               try:
                  it = iter(val)
               except Exception:
                  pass
               else:
                  good = True
                  try:
                     for x in val:
                        if (not isinstance(x, core.ImageBase) or
                            not val.data.pixel_type in self.pixel_types):
                           good = False
                           break
                  except Exception:
                     pass
                  else:
                     if good:
                        choices.append(key)
         else:
            for key, val in locals.items():
               if isinstance(val, core.ImageBase) and val.data.pixel_type in self.pixel_types:
                  choices.append(key)
      if self.has_default and isinstance(self.default,CNoneDefault):
          choices = ["None"] + choices
      return choices

class Rect(Class):
   pass

class Choice:
   def get_control(self, parent, locals=None):
      choices = []
      for choice in self.choices:
         if len(choice) == 2 and not util.is_string_or_unicode(choice):
            choices.append(choice[0])
         else:
            choices.append(choice)
      self.control = wx.Choice(parent, -1, choices=choices)
      if self.default < 0:
         self.default = len(choices) + self.default
      if self.default >= 0 and self.default < len(self.choices):
         self.control.SetSelection(self.default)
      return self

   def get_string(self):
      selection = self.control.GetSelection()
      if (len(self.choices[selection]) == 2 and
          not util.is_string_or_unicode(self.choices[selection])):
         return str(self.choices[selection][1])
      else:
         return str(selection)

   def get(self):
      selection = self.control.GetSelection()
      if (len(self.choices[selection]) == 2 and
          not util.is_string_or_unicode(self.choices[selection])):
         return self.choices[selection][1]
      else:
         return int(selection)

class ChoiceString:
   def get_control(self, parent, locals=None):
      if self.strict:
         self.control = wx.Choice(parent, -1, choices=self.choices)
         default_index = self.choices.index(self.default)
         self.control.SetSelection(default_index)
      else:
         if self.has_default:
            default = self.default
         else:
            default = self.choices[0]
         self.control = wx.ComboBox( 
            parent, -1, default, choices=self.choices, style=wx.CB_DROPDOWN)
      return self

   def get_string(self):
      return repr(self.get())

   def get(self):
      if self.strict:
         selection = self.control.GetStringSelection()
      else:
         selection = self.control.GetValue()
      return selection

class _Filename:
   def get_control(self, parent, locals=None, text=None):
      if text is None:
         text = self.default
      self.control = wx.BoxSizer(wx.HORIZONTAL)
      self.text = wx.TextCtrl(parent, -1, text, size=wx.Size(200, 24))
      browseID = wx.NewId()
      if wx.Platform == '__WXMAC__' and hasattr(buttons, 'ThemedGenButton'):
         browse = buttons.ThemedGenButton(
            parent, browseID, "...", size=wx.Size(24, 24))
      else:
         browse = wx.Button(
            parent, browseID, "...", size=wx.Size(24, 24))
      wx.EVT_BUTTON(browse, browseID, self.OnBrowse)
      self.control.Add(self.text, 1, wx.EXPAND)
      self.control.Add((4, 4), 0)
      self.control.Add(browse, 0)
      return self

   def get_string(self):
      text = self.text.GetValue()
      if text == "":
         return "None"
      else:
         return "r'" + text + "'"

   def get(self):
      text = self.text.GetValue()
      try:
         textutf8 = text.encode('utf8')
         text = textutf8
      except Exception:
         pass
      if text == "":
         return None
      else:
         return str(text)

class FileOpen(_Filename):
   def OnBrowse(self, event):
      parent = self.text.GetParent()
      filename = gui_util.open_file_dialog(parent, self.extension)
      if filename:
         self.text.SetValue(filename)
      parent.Raise()
      while wx.IsBusy():
         wx.EndBusyCursor()

   def get(self):
      while 1:
         text = self.text.GetValue()
         if not os.path.exists(os.path.abspath(text)):
            gui_util.message("File '%s' does not exist." % text)
            self.OnBrowse(None)
         else:
            break
      return _Filename.get(self)

   def get_string(self):
      while 1:
         text = self.text.GetValue()
         if not os.path.exists(os.path.abspath(text)):
            gui_util.message("File '%s' does not exist." % text)
            self.OnBrowse(None)
         else:
            break
      return _Filename.get_string(self)

class FileSave(_Filename):
   def OnBrowse(self, event):
      parent = self.text.GetParent()
      filename = gui_util.save_file_dialog(parent, self.extension)
      if filename:
         self.text.SetValue(filename)
      parent.Raise()
      while wx.IsBusy():
         wx.EndBusyCursor()

class Directory(_Filename):
   def OnBrowse(self, event):
      parent = self.text.GetParent()
      filename = gui_util.directory_dialog(self.text)
      if filename:
         self.text.SetValue(filename)
      parent.Raise()
      while wx.IsBusy():
         wx.EndBusyCursor()

   def get(self):
      while 1:
         text = self.text.GetValue()
         if not os.path.exists(os.path.abspath(text)):
            gui_util.message("File '%s' does not exist." % text)
            self.OnBrowse(None)
         else:
            break
      return _Filename.get(self)

   def get_string(self):
      while 1:
         text = self.text.GetValue()
         if not os.path.exists(os.path.abspath(text)):
            gui_util.message("File '%s' does not exist." % text)
            self.OnBrowse(None)
         else:
            break
      return _Filename.get_string(self)

class Radio:
   def get_control(self, parent, locals=None):
      self.control = wx.RadioButton(parent, -1, self.radio_button)
      return self

   def get_string(self):
      return str(self.control.GetValue())

   def get(self):
      return self.control.GetValue()

class Check:
   def get_control(self, parent, locals=None):
      self.control = wx.CheckBox(parent, -1, self.check_box)
      self.control.Enable(self.enabled)
      self.control.SetValue(self.default)
      return self

   def get_string(self):
      return str(self.control.GetValue())

   def get(self):
      return self.control.GetValue()

class Region(Class):
   def determine_choices(self, locals):
      from gamera import core
      self.klass = core.Region
      return Class.determine_choices(self, locals)

class RegionMap(Class):
   def determine_choices(self, locals):
      from gamera import core
      self.klass = core.RegionMap
      return Class.determine_choices(self, locals)

class ImageInfo(Class):
   def determine_choices(self, locals):
      from gamera import core
      self.klass = core.ImageInfo
      return Class.determine_choices(self, locals)

class ImageList(Class):
   def determine_choices(self, locals):
      from gamera import core
      self.klass = core.ImageBase
      return Class.determine_choices(self, locals)

class Info:
   def get_control(self, parent, locals=None):
      self.control = wx.StaticText(parent, -1, "")
      return self

   def get_string(self):
      return None
   get = get_string

class Pixel(_Filename):
   def get_control(self, parent, locals=None):
      if type(self.default) == RGBPixel:
         text = "RGBPixel(%d, %d, %d)" % (self.default.red,
                                          self.default.green,
                                          self.default.blue)
      else:
         text = str(self.default)
      return _Filename.get_control(self, parent, locals, text)

   def OnBrowse(self, event):
      dialog = wx.ColourDialog(None)
      if dialog.ShowModal() == wx.ID_OK:
         color = dialog.GetColourData().GetColour()
         self.text.SetValue("RGBPixel(%d, %d, %d)" % (color.Red(), color.Green(), color.Blue()))
      self.text.GetParent().Raise()

   def get_string(self):
      text = self.text.GetValue()
      if text == "":
         return "None"
      else:
         return text

   def get(self):
      text = self.text.GetValue()
      if text == "":
         return None
      else:
         return eval(text, {'RGBPixel': RGBPixel})

class PointVector:
   def determine_choices(self, locals):
      from gamera import core
      self.klass = core.Point
      return Class.determine_choices(self, locals)

class Point:
   def get_control(self, parent, locals=None):
      from gamera.core import Point
      default = Point(self.default)
      self.control = wx.BoxSizer(wx.HORIZONTAL)
      self.control.Add(wx.StaticText(parent, -1, "x:"))
      self.control_x = wx.SpinCtrl(
         parent, -1, value=str(default.x),
         min=-DEFAULT_MAX_ARG_NUMBER, max=DEFAULT_MAX_ARG_NUMBER,
         initial=default.x)
      self.control_x.SetValidator(_IntValidator(name=self.name))
      self.control.Add(self.control_x, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
      self.control.Add(wx.StaticText(parent, -1, "y:"))
      self.control_y = wx.SpinCtrl(
         parent, -1, value=str(default.y),
         min=-DEFAULT_MAX_ARG_NUMBER, max=DEFAULT_MAX_ARG_NUMBER,
         initial=default.y)
      self.control_y.SetValidator(_IntValidator(name=self.name))
      self.control.Add(self.control_y, 1, wx.EXPAND | wx.LEFT, 5)
      return self

   def get(self):
      from gamera.core import Point
      return Point(int(self.control_x.GetValue()),
                   int(self.control_y.GetValue()))
   
   def get_string(self):
      return str(self.get())

class Dim(Point):
   def get(self):
      from gamera.core import Dim
      return Dim(int(self.control_x.GetValue()),
                   int(self.control_y.GetValue()))
   
   def get_string(self):
      return str(self.get())

class FloatPoint:
   def get_control(self, parent, locals=None):
      from gamera.core import FloatPoint
      default = FloatPoint(self.default)
      self.control = wx.BoxSizer(wx.HORIZONTAL)
      self.control.Add(wx.StaticText(parent, -1, "x:"))
      self.control_x = wx.TextCtrl(
         parent, -1, str(default.x),
         validator = _RealValidator(name=self.name))
      self.control.Add(self.control_x, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
      self.control.Add(wx.StaticText(parent, -1, "y:"))
      self.control_y = wx.TextCtrl(
         parent, -1, str(default.y),
         validator = _RealValidator(name=self.name))
      self.control.Add(self.control_y, 1, wx.EXPAND | wx.LEFT, 5)
      return self

   def get(self):
      from gamera.core import FloatPoint
      return FloatPoint(float(self.control_x.GetValue()),
                        float(self.control_y.GetValue()))

   def get_string(self):
      return str(self.get())


from gamera import args
args.mixin(locals(), "GUI")



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import generators
import os, sys, dircache, glob, imp  # Python standard library

if 1:
   def dummy():
      pass

lib = os.path.dirname(os.path.realpath(dummy.func_code.co_filename))
lib_gui = os.path.realpath(os.path.join(lib, "gui"))
# Figure out if we are in the source directory or installed
plugins = os.path.realpath(os.path.join(lib, "plugins"))
doc = os.path.realpath(os.path.join(lib, "doc"))
sys.path.append(plugins)
plugins_src = ""
toolkits = os.path.realpath(os.path.join(lib, "toolkits"))
test = os.path.realpath(os.path.join(lib, "test"))
test_results = os.path.realpath(os.path.join(lib, "test/results"))

def get_toolkit_names(dir):
   toolkits = []
   listing = dircache.listdir(dir)
   dircache.annotate(dir, listing)
   for toolkit in listing:
      if toolkit.endswith(".py") and toolkit != "__init__.py":
         toolkits.append(toolkit[:-3])
      elif toolkit.endswith("module.so"):
         toolkits.append(toolkit[:-9])
      elif (toolkit.endswith("/") and
            "__init__.py" in dircache.listdir(os.path.join(dir, toolkit))):
         toolkits.append(toolkit[:-1])
   return toolkits

def get_directory_of_modules(dir, base=''):
   modules = glob.glob(os.path.join(dir, "*.py"))
   names = [os.path.basename(x).split('.')[0] for x in modules]
   mods = []
   suffixes = imp.get_suffixes()
   for i in suffixes:
      if i[0] == '.py':
         suffix = i
         break
   for m, name in zip(modules, names):
      try:
         module = imp.load_module(base + name, file(m, 'r'), m, suffix)
         mods.append(module)
      except Exception, e:
         print e
   return mods

def import_directory(dir, gl, lo, verbose=0):
   modules = glob.glob(os.path.join(dir, "*.py"))
   modules = [os.path.basename(x).split('.')[0] for x in modules]
   if verbose:
      sys.stdout.write("Loading plugins: " + "-" * 40 + "\n")
   column = 0
   first = 1
   result = []
     
   for m in modules:
      if m == '__init__':
         continue
      try:
         module = __import__(m, gl, lo, [])
         failed = 0
      except Exception, e:
         failed = e
      if failed:
         display = '[%s %s]' % (m, str(failed))
      else:
         display = m
         result.append(module)
      if m != modules[-1]:
         display += ", "
      column += len(display)
      if verbose:
         if column > 70:
            sys.stdout.write("\n")
            column = len(display)
         sys.stdout.write(display)
         sys.stdout.flush()
   if verbose:
      sys.stdout.write("\n")
   return result



import py.test

from gamera.core import *
init_gamera()


def test_load_image_rgb_rle():
   def _load_image_fail1():
      image = load_image("not_a_path")
   py.test.raises(IOError, _load_image_fail1)

def test_load_image_onebit():
   image = load_image("data/testline.tiff")
   assert image.pixel_type_name == "OneBit"
   assert image.nrows == 44
   assert image.ncols == 907
   assert image.black_area()[0] == 5174.0
   assert image.storage_format_name == "Dense"

   image2 = load_image("data/testline.png")
   assert image2.pixel_type_name == "OneBit"
   assert image2.nrows == 44
   assert image2.ncols == 907
   assert image2.black_area()[0] == 5174.0
   assert image.storage_format_name == "Dense"

   assert image.to_string() == image2.to_string()
   assert image.to_rle() == image2.to_rle()

def test_load_image_greyscale():
   greyscale = load_image("data/GreyScale_generic.tiff")
   assert greyscale.pixel_type_name == "GreyScale"
   assert greyscale.nrows == 68
   assert greyscale.ncols == 97
   assert greyscale.storage_format_name == "Dense"

   greyscale2 = load_image("data/GreyScale_generic.png")
   assert greyscale2.pixel_type_name == "GreyScale"
   assert greyscale2.nrows == 68
   assert greyscale2.ncols == 97
   assert greyscale2.storage_format_name == "Dense"

   assert greyscale.to_string() == greyscale2.to_string()

def test_load_image_greyscale_rle():
   def _load_image_greyscale_rle1():
      greyscale3 = load_image("data/GreyScale_generic.png", RLE)
   def _load_image_greyscale_rle2():
      greyscale3 = load_image("data/GreyScale_generic.tiff", RLE)
   py.test.raises(Exception, _load_image_greyscale_rle1)
   py.test.raises(Exception, _load_image_greyscale_rle2)

def test_load_image_rgb():
   rgb = load_image("data/RGB_generic.tiff")
   assert rgb.pixel_type_name == "RGB"
   assert rgb.nrows == 130
   assert rgb.ncols == 228
   assert rgb.storage_format_name == "Dense"

   rgb2 = load_image("data/RGB_generic.png")
   assert rgb2.pixel_type_name == "RGB"
   assert rgb2.nrows == 130
   assert rgb2.ncols == 228
   assert rgb.storage_format_name == "Dense"

   assert rgb.to_string() == rgb2.to_string()

def load_image_rgb_rle1():
   rgb3 = load_image("data/RGB_generic.png", RLE)

def load_image_rgb_rle2():
   rgb3 = load_image("data/RGB_generic.tiff", RLE)

def test_load_image_grey16():
   grey16 = load_image("data/Grey16_generic.tiff")
   assert grey16.pixel_type_name == "Grey16"
   assert grey16.nrows == 72
   assert grey16.ncols == 83
   assert grey16.storage_format_name == "Dense"

   grey16_2 = load_image("data/Grey16_generic.png")
   assert grey16_2.pixel_type_name == "Grey16"
   assert grey16_2.nrows == 72
   assert grey16_2.ncols == 83
   assert grey16.storage_format_name == "Dense"

   # assert grey16.to_string() == grey16_2.to_string()

def load_image_grey16_rle1():
   grey16 = load_image("data/Grey16_generic.png", RLE)

def load_image_grey16_rle2():
   grey16 = load_image("data/Grey16_generic.tiff", RLE)

# def test_load_image_grey16_rle():
#    py.test.raises(Exception, load_image_grey16_rle1)
#    py.test.raises(Exception, load_image_grey16_rle2)




#
# Copyright (C) 2012 Tobias Bolten
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx

from gamera.gui.gaoptimizer.ExpertSettingPanel import *

#-------------------------------------------------------------------------------
class StopCriteriaPanel(ExpertSettingPanel):
#-------------------------------------------------------------------------------
    #---------------------------------------------------------------------------
    def __init__(self, parent, id):
    #---------------------------------------------------------------------------
        ExpertSettingPanel.__init__(self, parent, id)

        sizer = wx.GridBagSizer(hgap=5, vgap=5)
        self.SetSizer(sizer)

        # best fitness
        self.bestFitness = wx.CheckBox(self, -1, "Perfect LOO-recognition reached", \
            name = "bestFitnessStop")
        sizer.Add(self.bestFitness, pos=(0,0), \
            flag = wx.LEFT | wx.RIGHT | wx.TOP | wx.EXPAND, border=10)

        self.genericWidgets.append(self.bestFitness)

        # generation counter
        self.maxGeneration = wx.CheckBox(self, -1, "Max. number of generations", \
            name = "maxGenerations")
        sizer.Add(self.maxGeneration, pos=(1,0), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border = 10)
        self.maxGenerationCount = wx.SpinCtrl(self, -1, size=(100,-1), \
            min=10, max=5000, value='100')
        self.maxGenerationCount.SetToolTipString("Number of generations")
        self.maxGenerationCount.Disable()
        sizer.Add(self.maxGenerationCount, pos=(1,1), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.genericWidgets.append(self.maxGeneration)
        self.AddChildToParent(self.maxGeneration, self.maxGenerationCount)

        # fitness counter
        self.maxFitnessEval = wx.CheckBox(self, -1, "Max. number of fitness evals", \
            name = "maxFitnessEvals")
        sizer.Add(self.maxFitnessEval, pos=(2,0), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)
        self.maxFitnessEvalCount = wx.SpinCtrl(self, -1, size=(100,-1), \
            min=10, max=50000, value='5000')
        self.maxFitnessEvalCount.SetToolTipString("Number of evaluations")
        self.maxFitnessEvalCount.Disable()
        sizer.Add(self.maxFitnessEvalCount, pos=(2,1), \
            flag = wx.LEFT | wx.RIGHT | wx.EXPAND, border=10)

        self.genericWidgets.append(self.maxFitnessEval)
        self.AddChildToParent(self.maxFitnessEval, self.maxFitnessEvalCount)

        # steady state continue
        self.steadyContinue = wx.CheckBox(self, -1, "Steady state continue", \
            name = "steadyStateStop")
        self.steadyContinue.SetValue(True)
        sizer.Add(self.steadyContinue, pos=(3,0), \
            flag = wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, border=10)
        self.steadyContinueMin = wx.SpinCtrl(self, -1, size=(100,-1), \
            min=10, max=250000, value='40')
        self.steadyContinueMin.SetToolTipString("Minimum generations")
        sizer.Add(self.steadyContinueMin, pos=(3,1), \
            flag = wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, border=10)
        self.steadyContinueNoChange = wx.SpinCtrl(self, -1, size=(100,-1), \
            min=1, max=10000, value='10')
        self.steadyContinueNoChange.SetToolTipString("Generations without improvement")
        sizer.Add(self.steadyContinueNoChange, pos=(3,2), \
            flag = wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, border=10)

        self.genericWidgets.append(self.steadyContinue)
        self.AddChildToParent(self.steadyContinue, self.steadyContinueMin)
        self.AddChildToParent(self.steadyContinue, self.steadyContinueNoChange)

        # bind the EVT_CHECKBOX to the CheckBoxes
        self.BindEvent(wx.EVT_CHECKBOX, self.OnCheckBox, \
            [self.bestFitness, self.maxGeneration,
             self.maxFitnessEval, 
             self.steadyContinue])



#
# This file has been taken from wxpython (see the file
# wx/tools/img2img.py in the wxpython source distribution)
#
# Copyright (c) 1998 Julian Smart, Robert Roebling et al
#
# This program may be freely used, copied and distributed under
# the terms of the wxWindows Library Licence, Version 3. See
# the file "copyright" of the wxpython distribution from
# http://wxpython.org/ for details.
#
"""
Common routines for the image converter utilities.
"""
import sys, os, glob, getopt, string
import wx

if wx.Platform == "__WXGTK__":
    # some bitmap related things need to have a wxApp initialized...
    app = wx.PySimpleApp()

wx.InitAllImageHandlers()

def convert(file, maskClr, outputDir, outputName, outType, outExt):
    if string.lower(os.path.splitext(file)[1]) == ".ico":
        icon = wx.Icon(file, wx.BITMAP_TYPE_ICO)
        img = wx.BitmapFromIcon(icon)
    else:
        img = wx.Bitmap(file, wx.BITMAP_TYPE_ANY)

    if not img.Ok():
        return 0, file + " failed to load!"
    else:
        if maskClr:
            om = img.GetMask()
            mask = wx.MaskColour(img, maskClr)
            img.SetMask(mask)
            if om is not None:
                om.Destroy()
        if outputName:
            newname = outputName
        else:
            newname = os.path.join(outputDir,
                                   os.path.basename(os.path.splitext(file)[0]) + outExt)
        if img.SaveFile(newname, outType):
            return 1, file + " converted to " + newname
        else:
            img = wx.ImageFromBitmap(img)
            if img.SaveFile(newname, outType):
                return 1, "ok"
            else:
                return 0, file + " failed to save!"




def main(args, outType, outExt, doc):
    if not args or ("-h" in args):
        print doc
        return

    outputDir = ""
    maskClr = None
    outputName = None

    try:
        opts, fileArgs = getopt.getopt(args, "m:n:o:")
    except getopt.GetoptError:
        print __doc__
        return

    for opt, val in opts:
        if opt == "-m":
            maskClr = val
        elif opt == "-n":
            outputName = val
        elif opt == "-o":
            outputDir = val

    if not fileArgs:
        print doc
        return

    for arg in fileArgs:
        for file in glob.glob(arg):
            if not os.path.isfile(file):
                continue
            ok, msg = convert(file, maskClr, outputDir, outputName,
                              outType, outExt)
            print msg




"""Data transfer between Gamera and OpenCV
"""

from gamera.plugin import *
from gamera import config

try:
    import cv
except ImportError:
    try:
        verbose = config.get("verbosity_level")
    except Exception:
        verbose = 0
    if verbose:
        print ('Info: OpenCV Library cv '
               'could not be imported')
    
class from_cv(PluginFunction):
    """
    Instantiates a Gamera RGB image from an OpenCV
    cv.cvmat or cv.iplimage. May fail for very large images.

    Usage example:

    .. code:: Python

      from gamera.plugins import cv_io

      # read a JPEG image and convert it to a Gamera image
      cvImg = cv.LoadImage("foo.jpg")
      img = cv_io.from_cv(cvImg)
    """
    self_type = None
    return_type = ImageType([GREYSCALE, RGB, FLOAT])
    args = Args([Class("image")])
    pure_python = True

    def __call__(image, offset=None):
        from gamera.plugins import _string_io
        from gamera.core import Dim, Point, RGBPixel
        
        if offset is None:
          offset = Point(0, 0)
        if isinstance(image,cv.cvmat):
          imgcv = cv.CloneMat(image)
          cv.CvtColor(image, imgcv, cv.CV_BGR2RGB)
          return  _string_io._from_raw_string(
            offset, 
            Dim(imgcv.cols,imgcv.rows), 
            RGB, DENSE, 
            imgcv.tostring())
        elif isinstance(image,cv.iplimage):
          imgcv = cv.CreateImage(cv.GetSize(image), image.depth, image.channels)
          cv.CvtColor(image, imgcv, cv.CV_BGR2RGB)
          return _string_io._from_raw_string(
            offset, 
            Dim(imgcv.width,imgcv.height), 
            RGB, DENSE, 
            imgcv.tostring())
        else:
          raise TypeError("Image must be of type cv.cvmat or cv.iplimage")
          
    __call__ = staticmethod(__call__)

class to_cv(PluginFunction):
    """
    Returns an OpenCv image (cv.cvmat) containing a copy of
    image's data.

    Only RGB, Greyscale and Onebit images are supported.
    May fail for very large images.
    """
    self_type = ImageType([RGB, GREYSCALE, ONEBIT])
    return_type = Class("cv_image")
    def __call__(image):
        from gamera.plugins import _string_io
        
        if image.data.pixel_type == ONEBIT:
          image = image.to_rgb()
        if image.data.pixel_type == GREYSCALE:
          image = image.to_rgb()
        numImage = image.to_numeric()
        imgcv = cv.fromarray(numImage)
        cv.CvtColor(imgcv, imgcv, cv.CV_BGR2RGB)
        return imgcv
        
    __call__ = staticmethod(__call__)
  

class CVioModule(PluginModule):
    category = "ExternalLibraries/cv"
    author = "Manuel Jeltsch"
    functions = [from_cv, to_cv]
    pure_python = True
    url = ('')
module = CVioModule()

from_cv = from_cv()



from gamera.core import *
from gamera import knn, classify, gamera_xml
init_gamera()

correct_classes = ['latin.lower.letter.h', 'latin.lower.ligature.ft', 'latin.capital.letter.m', '_group._part.latin.capital.letter.m', '_group._part.latin.capital.letter.m', '_group._part.latin.lower.letter.i', 'latin.lower.letter.d', 'latin.capital.letter.m', 'latin.capital.letter.t', '_group._part.latin.lower.letter.h', 'latin.lower.ligature.fi', '_group._part.latin.lower.ligature.ft', '_group._part.latin.lower.letter.h', '_group._part.latin.lower.letter.i', 'latin.lower.letter.h', 'latin.lower.letter.d', 'latin.lower.letter.d', 'latin.capital.letter.m', 'latin.capital.letter.c', 'latin.lower.letter.t', 'latin.lower.letter.t', '_group._part.latin.lower.letter.n', 'latin.lower.letter.e', 'latin.lower.letter.a', 'latin.lower.letter.r', 'latin.lower.letter.r', 'latin.lower.letter.a', '_group._part.latin.lower.letter.n', '_group._part.latin.lower.letter.i', 'latin.lower.letter.a', 'latin.lower.letter.r', 'latin.lower.letter.r', '_group._part.latin.lower.letter.h', 'latin.lower.letter.e', 'latin.lower.letter.r', 'latin.lower.letter.e', 'latin.lower.letter.n', 'latin.lower.letter.n', 'latin.lower.letter.o', 'latin.lower.letter.e', 'latin.lower.letter.s', 'latin.lower.letter.e', '_group._part.latin.lower.letter.h', '_group._part.latin.lower.letter.i', '_group._part.latin.lower.letter.g', 'latin.lower.letter.a', 'latin.lower.letter.r', 'latin.lower.letter.r', 'latin.lower.letter.o-', 'latin.lower.letter.r', 'hyphen-minus', 'comma', 'full.stop', 'comma', '_group._part.latin.lower.ligature.ft', 'noise', '_group._part.latin.lower.letter.g']


results = [
      ['latin.lower.letter.n', 'latin.capital.letter.m', 'latin.lower.letter.h', 'latin.lower.ligature.ft', 'latin.capital.letter.t', 'latin.lower.letter.g'],
      ['latin.lower.letter.n', 'latin.capital.letter.m', 'latin.lower.letter.h', 'latin.lower.ligature.ft', 'latin.lower.letter.h', 'latin.lower.letter.g'],
      ['latin.lower.letter.n', 'latin.capital.letter.m', 'latin.capital.letter.t', 'latin.lower.letter.h', 'latin.lower.ligature.ft', 'latin.lower.letter.h', 'latin.lower.letter.i', 'latin.lower.letter.g'],
      ['latin.lower.letter.n', 'latin.capital.letter.m', 'latin.capital.letter.t', 'latin.lower.letter.h', 'latin.lower.ligature.ft', 'latin.lower.letter.h', 'latin.lower.letter.i', 'latin.lower.letter.g']
   ]

#featureset = ['area', 'aspect_ratio', 'black_area', 'compactness', 'moments', 'ncols_feature', 'nholes', 'nholes_extended', 'nrows_feature', 'skeleton_features', 'top_bottom', 'volume', 'volume16regions', 'volume64regions', 'zernike_moments']
featureset = ['area', 'aspect_ratio', 'black_area', 'moments', 'nholes_extended', 'skeleton_features', 'volume64regions']

def _test_grouping(classifier, ccs):
   classifier.change_feature_set(featureset)
   cases = [(classify.ShapedGroupingFunction(4), 'min'),
             (classify.ShapedGroupingFunction(4), 'avg'),
             (None, 'min'),
             (None, 'avg')
            ]

   for (i, (func, criterion)) in enumerate(cases):
        if func == None:
            added,removed = classifier.group_list_automatic(ccs, criterion=criterion)
        else:
            added, removed = classifier.group_list_automatic(
               ccs,
               grouping_function = func,
               max_parts_per_group = 10,
               max_graph_size = 64,
               criterion=criterion)
   
        added.sort(lambda a,b: cmp(a.offset_x,b.offset_x))
        assert [cc.get_main_id() for cc in added] == results[i]


def _test_classification(classifier, ccs):
   (id_name, confidence) = classifier.guess_glyph_automatic(ccs[0])
   assert id_name == [(1.0, 'latin.lower.letter.h')]

   classifier.classify_glyph_automatic(ccs[1])
   assert ccs[1].id_name == [(1.0, 'latin.lower.ligature.ft')]
   
   added, removed = classifier.classify_list_automatic(ccs)
   assert [cc.get_main_id() for cc in ccs] == correct_classes
   assert added == [] and removed == []

   classifier.classify_and_update_list_automatic(ccs)
   assert [cc.get_main_id() for cc in ccs] == correct_classes

   classifier.change_feature_set(['area'])
   assert len(list(classifier.database)[0].features) == 1

   added, removed = classifier.classify_list_automatic(ccs)
   assert [cc.get_main_id() for cc in ccs] != correct_classes

   _test_grouping(classifier, ccs)


def _test_training(classifier, ccs):
   length = len(classifier.get_glyphs())
   classifier.classify_glyph_manual(ccs[0], "dummy")
   assert len(classifier.get_glyphs()) == length + 1
   added, removed = classifier.classify_list_manual(ccs, "dummy")
   assert len(classifier.get_glyphs()) == length + len(ccs)
   assert added == [] and removed == []
   classifier.classify_and_update_list_manual(ccs, "dummy")
   assert len(classifier.get_glyphs()) == length + len(ccs)
   classifier.add_to_database(ccs)
   assert len(classifier.get_glyphs()) == length + len(ccs)
   classifier.remove_from_database(ccs)
   assert len(classifier.get_glyphs()) == length
   classifier.add_to_database(ccs)
   assert len(classifier.get_glyphs()) == length + len(ccs)

def test_interactive_classifier():
   # We assume the XML reading/writing itself is fine (given
   # test_xml), but we should test the wrappers in classify anyway
   image = load_image("data/testline.png")
   ccs = image.cc_analysis()

   classifier = knn.kNNInteractive([],features=featureset)
   assert classifier.is_interactive()
   assert len(classifier.get_glyphs()) == 0
   
   classifier.from_xml_filename("data/testline.xml")
   assert len(classifier.get_glyphs()) == 66
   _test_classification(classifier, ccs)
   _test_training(classifier, ccs)
   length = len(classifier.get_glyphs())

   # subtract len(group_parts) because to_xml_filename() does 
   # not save "_group._part"
   group_parts = [x for x in classifier.get_glyphs() 
                  if x.get_main_id().startswith("_group._part")]
   length = length - len(group_parts)

   classifier.to_xml_filename("tmp/testline_classifier.xml")
   classifier.from_xml_filename("tmp/testline_classifier.xml")
   assert len(classifier.get_glyphs()) == length
   classifier.merge_from_xml_filename("data/testline.xml")
   assert len(classifier.get_glyphs()) == length + 66
   classifier.clear_glyphs()
   assert len(classifier.get_glyphs()) == 0
   classifier.from_xml_filename("data/testline.xml")
   assert len(classifier.get_glyphs()) == 66
   
def test_noninteractive_classifier():
   # We assume the XML reading/writing itself is fine (given
   # test_xml), but we should test the wrappers in classify anyway
   image = load_image("data/testline.png")
   ccs = image.cc_analysis()

   database = gamera_xml.glyphs_from_xml("data/testline.xml")
   classifier = knn.kNNNonInteractive(database,features=featureset,normalize=False)
   assert not classifier.is_interactive()
   assert len(classifier.get_glyphs()) == 66
   
   _test_classification(classifier, ccs)

   classifier.serialize("tmp/serialized.knn")
   classifier.clear_glyphs()
   assert len(classifier.get_glyphs()) == 0
   classifier.unserialize("tmp/serialized.knn")




#
# Copyright (C) 2001-2005 Michael Droettboom and Robert Ferguson
#               2009      Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Edge detectors based on first and second derivatives, and related
post-processing from the VIGRA Computer Vision Library."""

from gamera.plugin import *
import _edgedetect

########################################
# Edge detection methods

class difference_of_exponential_edge_image(PluginFunction):
      u"""
      EXPERIMENTAL

      Detect and mark edges in an edge image using the Shen/Castan
      zero-crossing detector.

      Uses code from the VIGRA Computer Vision Library (Copyright
      1998-2007 by Ullrich K\u00f6the).

      *scale*
        The scale relates to the value b of the exponential filter.

      *gradient_threshold*
	Whenever the gradient at a zero crossing is greater than the
	given gradient_threshold, an edge point is marked in the
	destination image on the darker side of the zero crossing (the
	zero crossing occurs between pixels).

      *min_edge_length*
	Removes all edges shorter than the number of pixels specified
	(0 retains all edges).  Values near 10 are suggested.
      """
      self_type = ImageType([GREYSCALE, GREY16, FLOAT])
      args = Args([Real("scale", (0, 1e300), default=0.8),
                   Real("gradient_threshold", (0, 1e300), default=4.0),
                   Int("min_edge_length", (0, 32000), default=0)])
      return_type = ImageType([GREYSCALE, GREY16, FLOAT])
      def __call__(self, scale=0.8, gradient_threshold=4.0, min_edge_length=0):
            return _edgedetect.difference_of_exponential_edge_image(
                  self, scale, gradient_threshold, min_edge_length)
      __call__ = staticmethod(__call__)
      doc_examples = [(GREYSCALE,)]

class difference_of_exponential_crack_edge_image(PluginFunction):
      u"""EXPERIMENTAL

      Detect and mark edges in a crack edge image using the
      Shen/Castan zero-crossing detector.

      Uses code from the VIGRA Computer Vision Library (Copyright
      1998-2007 by Ullrich K\u00f6the).

      *scale*
	The scale relates to the value b of the exponential filter.

      *gradient_threshold*
	Whenever the gradient at a zero crossing is greater than the
	given gradient threshold, an edge point is marked in the
	destination image on the darker side of the zero crossing (the
	zero crossing occurs between pixels).

      *min_edge_length*
	Removes all edges shorter than the number of pixels specified
	(0 retains all edges).  Values near 10 are suggested.

      *close_gaps*
        Close one pixel wide gaps.

      *beautify*
        See the VIGRA Docs.
      """
      self_type = ImageType([GREYSCALE, GREY16, FLOAT])
      args = Args([Real("scale", (0, 1e300), default=0.8),
                   Real("gradient_threshold", (0, 1e300), default=4.0),
                   Int("min_edge_length", (0, 32000), default=0),
                   Check("close_gaps", default=False), 
                   Check("beautify", default=False)])
      return_type = ImageType([GREYSCALE, GREY16, FLOAT])
      def __call__(self, scale=0.8, gradient_threshold=4.0, min_edge_length=0,
                   close_gaps=False, beautify=False):
            return _edgedetect.difference_of_exponential_crack_edge_image(
                  self, scale,
                  gradient_threshold, min_edge_length, close_gaps, beautify)
      __call__ = staticmethod(__call__)
      doc_examples = [(GREYSCALE,)]

class canny_edge_image(PluginFunction):
      u"""EXPERIMENTAL

      Detect and mark edges in an edge image using Canny's algorithm.

      Uses code from the VIGRA Computer Vision Library (Copyright
      1998-2007 by Ullrich K\u00f6the).

      *scale*
	The scale relates to the value b of the exponential filter.

      *gradient_threshold*
	This operator first calls cannyEdgelList() to generate an
	edgel list for the given image. Than it scans this list and
	selects edgels whose strength is above the given
	gradient_threshold.
      """
      self_type = ImageType([GREYSCALE, GREY16, FLOAT])
      args = Args([Real("scale", [0, 1e300], default=0.8),
                   Real("gradient_threshold", [0, 1e300], default=4.0)])
      return_type = ImageType([GREYSCALE, GREY16, FLOAT])
      def __call__(self, scale=0.8, gradient_threshold=4.0):
            return _edgedetect.canny_edge_image(self, scale, gradient_threshold)
      __call__ = staticmethod(__call__)
      doc_examples = [(GREYSCALE,)]

class labeled_region_edges(PluginFunction):
  """
  Pixels with a label different from one of its neighboring pixels
  are marked black in the returned image.

  When *mark_both* is ``True``, both edges of the region border are
  marked, resulting in a two pixel wide edge.
  """
  self_type = ImageType([ONEBIT, GREYSCALE, GREY16, RGB])
  args = Args([Check("mark_both", default=False)])
  return_type = ImageType([ONEBIT])
  author = "Christoph Dalitz"
  # wrapper for passing default argument
  def __call__(self, mark_both=False):
      return _edgedetect.labeled_region_edges(self, mark_both)
  __call__ = staticmethod(__call__)

class outline(PluginFunction):
    """
    Traces the outline of the image.  When *which* is 0 (\"outer\"),
    the result is obtained by dilating the image and then XOR'ing the
    result with the original; when *which* is 1 (\"inner\"), the result
    is obtained by eroding the image and then XOR'ing the
    result with the original.
    """
    self_type = ImageType([ONEBIT])
    args = Args([Choice("which", ["outer", "inner"], default=0)])
    return_type = ImageType([ONEBIT])
    doc_examples = [(ONEBIT, 0), (ONEBIT, 1)]
    def __call__(self, which=0): 	
        return _edgedetect.outline(self, which)
    __call__ = staticmethod(__call__)


class EdgeDetect(PluginModule):
      category = "Edge"
      cpp_headers=["edgedetect.hpp"]
      functions = [difference_of_exponential_edge_image,
                   difference_of_exponential_crack_edge_image,
                   canny_edge_image,
                   labeled_region_edges,
                   outline]
      author = u"Ullrich K\u00f6the (wrapped by Robert Ferguson)"
      url = "http://gamera.sourceforge.net/"
module = EdgeDetect()




from gamera.core import *
init_gamera()

def test_rle1():
   image1 = load_image("data/testline.png")
   image2 = load_image("data/testline.png", RLE)

   # Check basic RLE image loading
   assert image2.pixel_type_name == "OneBit"
   assert image2.nrows == 44
   assert image2.ncols == 907
   assert image2.black_area()[0] == 5174.0

   # Compare RLE to DENSE image
   assert image1._to_raw_string() == image2._to_raw_string()
   assert image1.to_rle() == image2.to_rle()

   # Run something complicated, like cc_analysis
   image1.cc_analysis()
   image2.cc_analysis()

   # compare the results
   assert image1.color_ccs().to_string() == image2.color_ccs().to_string()
   assert image1._to_raw_string() == image2._to_raw_string()

   # Do some more complicated stuff, particularly things that
   # use iterators in unusual ways
   assert image1.projection_rows() == image2.projection_rows()
   assert image1.projection_cols() == image2.projection_cols()
   assert image1.most_frequent_run("black","vertical") == image2.most_frequent_run("black","vertical")
   assert image1.most_frequent_run("black","horizontal") == image2.most_frequent_run("black","horizontal")

   



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2010 Ichiro Fujinaga, Michael Droettboom,
#                         Karl MacMillan, Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.gui import has_gui
from new import instancemethod
import sys, os.path   # Python standard library
from types import *
import util, paths            # Gamera specific

# the maximum default number (int, float) in argument boxes
# necessary because sys.maxint is not accepted by some wxPython widgets
DEFAULT_MAX_ARG_NUMBER = 1048576    # 2^20

######################################################################

# This is a "self-generating" dialog box

class NoGUIArgs:
   def setup(self, *args, **kwargs):
      raise Exception("No GUI environment available.  Cannot display dialog.")
   
   def show(self, *args, **kwargs):
      raise Exception("No GUI environment available.  Cannot display dialog.")

class Args(NoGUIArgs):
   # list is a list of "Arg s"
   def __init__(self, list=[], name="Arguments", function=None, title=None):
      self.list = util.make_sequence(list)
      self.valid = 1
      self.name = name
      self.function = function
      self.title = title

   def __repr__(self):
      return "<" + self.__class__.__name__ + ">"

   def __getitem__(self, i):
      return self.list[i]
   index = __getitem__

   def __len__(self, i):
      return len(self.list)

######################################################################

# ARGUMENT TYPES

# for allowing None as default argument
class CNoneDefault:
   def __str__(self):
      return "None"
NoneDefault = CNoneDefault()

class Arg:
   default = 0
   length = 1

   def __init__(self, name):
      self.name = name
      if name is not None and not util.is_string_or_unicode(name):
         raise TypeError("'name' must be a string")
      self.has_default = False

   def __repr__(self):
      return "<" + self.__class__.__name__ + ">"

   def rest_repr(self, name=False):
      result = "``%s``" % self.__class__.__name__
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % str(self.default)
      return result

   def register(self, plug, func):
      pass
   
class Int(Arg):
   def __init__(self, name=None, range=(-DEFAULT_MAX_ARG_NUMBER, DEFAULT_MAX_ARG_NUMBER), default=None):
      Arg.__init__(self, name)
      if not (util.is_sequence(range) and len(range) == 2 and
              type(range[0]) in (int, float) and type(range[1]) in (int, float)):
         raise TypeError("'range' must be a 2-tuple of numbers")
      self.rng = range
      if default is None:
         self.has_default = False
         self.default = 0
      else:
         self.has_default = True
         self.default = default
      if not isinstance(self.default,CNoneDefault) and type(self.default) != int:
         raise TypeError("'default' must be an int")

   def rest_repr(self, name=False):
      result = "int"
      if self.rng != (-DEFAULT_MAX_ARG_NUMBER, DEFAULT_MAX_ARG_NUMBER):
         result += "(%d, %d)" % tuple([int(x) for x in self.rng])
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            if isinstance(self.default,CNoneDefault) or isinstance(self.default, NoneType):
                result += " = %s" % str(self.default)
            else:
                result += " = %d" % self.default
      return result

class Real(Arg):
   def __init__(self, name=None,
                range=None,
                default=None):
      Arg.__init__(self, name)
      if range is None:
         range = (-DEFAULT_MAX_ARG_NUMBER, DEFAULT_MAX_ARG_NUMBER)
      if not (util.is_sequence(range) and len(range) == 2 and
              type(range[0]) in (int, float) and type(range[1]) in (int, float)):
         raise TypeError("'range' must be a 2-tuple of numbers")
      self.rng = range
      if default is None:
         self.has_default = False
         self.default = 0.0
      else:
         self.has_default = True
         self.default = default
      if type(self.default) != float:
         raise TypeError("'default' must be a float")

   def rest_repr(self, name=False):
      result = "float"
      if self.rng != (-DEFAULT_MAX_ARG_NUMBER, DEFAULT_MAX_ARG_NUMBER):
         result += "(%.02f, %.02f)" % tuple([float(x) for x in self.rng])
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %.02f" % self.default
      return result

Float = Real

class Complex(Arg):
   def __init__(self, name=None, default=None):
      Arg.__init__(self, name)
      if default is None:
         self.has_default = False
         self.default = 0j
      else:
         self.has_default = True
         self.default = default

   def rest_repr(self, name=False):
      result = "complex"
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % self.default
      return result

class String(Arg):
   def __init__(self, name=None, default=None):
      Arg.__init__(self, name)
      if default is None:
         self.has_default = False
         self.default = ''
      else:
         self.has_default = True
         self.default = default
      if not util.is_string_or_unicode(self.default):
         raise TypeError("'default' must be an int")

   def rest_repr(self, name=False):
      result = "str"
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % repr(self.default)
      return result

class Class(Arg):
   def __init__(self, name=None, klass=None, list_of=False, default=None):
      Arg.__init__(self, name)
      self.klass = klass
      self.list_of = bool(list_of)
      if default is None:
         self.has_default = False
         self.default = ''
      else:
         self.has_default = True
         self.default = default

   def rest_repr(self, name=False):
      if self.klass is None:
         result = "object"
      else:
         result = "%s" % self.klass.__name__
      if name:
         result += " *%s*" % self.name
      if self.list_of:
         result = "[%s]" % result
      if self.has_default:
         result += " = %s" % str(self.default)
      return result

class ImageType(Arg):
   def __init__(self, pixel_types, name=None, list_of=False, default=None):
      import core
      Arg.__init__(self, name)
      if not util.is_sequence(pixel_types):
         pixel_types = (pixel_types,)
      if not util.is_homogenous_list(pixel_types, (int,)):
         raise TypeError("'pixel_types' must be a list of integers.")
      if not core is None:
         self.klass = core.ImageBase
      else:
         self.klass = None
      self.pixel_types = pixel_types
      self.list_of = bool(list_of)
      if default is None:
         self.has_default = False
         self.default = None
      else:
         self.has_default = True
         self.default = default
      if self.has_default and not isinstance(self.default,CNoneDefault):
         raise TypeError("'default' can only be NoneDefault")

   def rest_repr(self, name=False):
      result = '``Image`` [%s]' % '|'.join([util.get_pixel_type_name(x) for x in self.pixel_types])
      if name:
         result += " *%s*" % self.name
      if self.list_of:
         result = "[%s]" % result
      if self.has_default:
          result += " = %s" % str(self.default)
      return result

   def register(self, plug, func):
      from gamera import core
      func = instancemethod(func, None, core.gameracore.Image)
      setattr(core.ImageBase, plug.__name__, func)

class Rect(Arg):
   def __init__(self, name=None, list_of=False):
      import core
      Arg.__init__(self, name)
      if not core is None:
         self.klass = core.Rect
      self.list_of = bool(list_of)
   
   def rest_repr(self, name=False):
      result = Arg.rest_repr(self, name)
      if self.list_of:
         return "[%s]" % result
      return result
      
class Choice(Arg):
   def __init__(self, name=None, choices=[], default=None):
      Arg.__init__(self, name)
      if not util.is_string_or_unicode_list(choices):
         raise TypeError("'choices' must be a list of strings.")
      self.choices = choices
      if default is None:
         self.has_default = False
         self.default = 0
      else:
         self.has_default = True
         self.default = default
      if not isinstance(self.default,CNoneDefault) and type(self.default) != int:
         raise TypeError("'default' must be an int")

   def rest_repr(self, name = False):
      result = '``Choice`` [%s]' % ('|'.join(self.choices))
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            if isinstance(self.default,CNoneDefault):
               result += " = %s" % str(self.default)
            else:
               result += " = %s" % self.choices[self.default]
      return result

class ChoiceString(Arg):
   def __init__(self, name=None, choices=[], default=None, strict=True):
      Arg.__init__(self, name)
      if not util.is_string_or_unicode_list(choices):
         raise TypeError("'choices' must be a list of strings.")
      self.choices = choices
      if default is None:
         self.has_default = False
         self.default = choices[0]
      else:
         if type(default) != str:
            raise TypeError("'default' must be a string")
         if default not in choices:
            raise ValueError("Given 'default' must be in available 'choices'")
         self.has_default = True
         self.default = default
      self.strict = strict

   def rest_repr(self, name = False):
      if self.strict:
         strict = "(strict)"
      else:
         strict = ""
      result = '``ChoiceString%s`` [%s]' % (strict, '|'.join(self.choices))
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % self.default
      return result

class _Filename(Arg):
   def __init__(self, name=None, default="", extension="*.*"):
      Arg.__init__(self, name)
      if not util.is_string_or_unicode(default):
         raise TypeError("'default' must be a string")
      self.default = default
##       if not util.is_string_or_unicode(extension):
##          raise TypeError("'extension' must be a string")
      self.extension = extension

class FileOpen(_Filename):
   pass

class FileSave(_Filename):
   pass

class Directory(_Filename):
   pass

class Radio(Arg):
   def __init__(self, name=None, radio_button=''):
      Arg.__init__(self, name)
      self.radio_button = radio_button

class Check(Arg):
   def __init__(self, name=None, check_box='', default=None, enabled=True):
      Arg.__init__(self, name)
      if not util.is_string_or_unicode(check_box):
         raise TypeError("'check_box' must be a string")
      self.check_box = check_box
      if default is None:
         self.has_default = False
         self.default = False
      else:
         self.has_default = True
         self.default = default
      self.default = bool(self.default)
      self.enabled = bool(enabled)

   def rest_repr(self, name=False):
      result = '``bool``'
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % str(self.default)
      return result

Bool = Check

class Region(Class):
   def __init__(self, name=None):
      Class.__init__(self, name, None)

class RegionMap(Class):
   def __init__(self, name=None):
      Class.__init__(self, name, None)

class ImageInfo(Class):
   def __init__(self, name=None):
      Class.__init__(self, name, None)

class Point(Arg):
   def __init__(self, name=None, default=None):
      Arg.__init__(self, name)
      if default is None:
         self.has_default = False
         self.default = (0, 0)
      else:
         self.has_default = True
         self.default = default

class FloatPoint(Arg):
   def __init__(self, name=None, default=None):
      Arg.__init__(self, name)
      if default is None:
         self.has_default = False
         self.default = (0.0, 0.0)
      else:
         self.has_default = True
         self.default = default

class Dim(Point):
   pass

class _Vector(Class):
   def __init__(self, name=None, default=None, length=-1):
      Class.__init__(self, name, self.klass)
      if type(length) != int:
         raise TypeError("'length' must be an int")
      self.length = length
      if default is None:
         self.default = []
         self.has_default = False
      else:
         self.default = default
         self.has_default = True

   def rest_repr(self, name=False):
      result = "``%s``" % self.__class__.__name__
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % str(self.default)
      return result

class FloatVector(_Vector):
   klass = float
   typecode = "d"

class IntVector(_Vector):
   klass = int
   typecode = "i"

class ComplexVector(_Vector):
   klass = complex
   typecode = None

class ImageList(Class):
   def __init__(self, name=None, default=None):
      Class.__init__(self, name=name, klass=None, list_of=True, default=default)

class Pixel(Class):
   """Uses these to pass pixel values into a plugin.  The
PixelType must always match the type of the "self" image."""
   def __init__(self, name=None, default=None):
      Class.__init__(self, name)
      if default is None:
         self.has_default = False
      else:
         self.has_default = True
      self.default = default

   def rest_repr(self, name=False):
      result = "Pixel"
      if name:
         result += " *%s*" % self.name
         if self.has_default:
            result += " = %s" % str(self.default)
      return result

class Info(Arg):
   pass

class PointVector(Class):
   def __init__(self, name=None):
      Class.__init__(self, name, None, True)

class Wizard:
   def show(self, dialog):
      dialog_history = ['start', dialog]
      next_dialog = dialog
      while next_dialog != None:
         if next_dialog == 'start':
            return
         result = next_dialog.show(self.parent, self.locals, wizard=1)
         if not result is None:
            next_dialog = getattr(self, next_dialog.function)(*next_dialog.get_args())
            if next_dialog != dialog_history[-1]:
               dialog_history.append(next_dialog)
         else:
            next_dialog = dialog_history[-2]
            dialog_history = dialog_history[0:-1]
      self.done()

__all__ = 'Args Int Real Float Complex String Class ImageType Rect Choice FileOpen FileSave Directory Radio Check Region RegionMap ImageInfo FloatVector IntVector ComplexVector ImageList Info Wizard Pixel PointVector _Vector ChoiceString Point FloatPoint Dim Bool'.split()

___mixin_locals = locals()
def mixin(module, name):
   for cls_name in __all__ + ["Arg"]:
      cls = ___mixin_locals[cls_name]
      if module.has_key(cls_name):
         cls.__bases__ = tuple([module[cls_name]] + list(cls.__bases__))
   sys.stdout.write("\n")



import py.test

from gamera.core import *
init_gamera()

def make_test(inner):
   def _generate_rgb():
      rng = range(0, 255, 8)
      for r in rng:
         for g in rng:
            for b in rng:
               yield RGBPixel(r,g,b)

   def test():
      for type, value in [(ONEBIT, xrange((2 ** 16) - 1)),
                          (GREYSCALE, xrange((2 ** 8) - 1)),
                          (FLOAT, [float(x) for x in xrange(-5, 5)]),
                          (COMPLEX, [complex(x) for x in xrange(-5, 5)]),
                          (RGB, _generate_rgb()),
                          (GREY16, xrange((2 ** 16) - 1))]:
         inner(type, value, DENSE)
      inner(ONEBIT, xrange(0, 2 ** 16 - 1), RLE)
   return test

def _test_image_constructors(type, value, storage):
   def _fail1():
      Image(Size(50, 50), Point(25, 25), type, storage)
   #image = Image(25, 25, 50, 50, type, storage) # deprecated call
   image = Image(Point(25,25), Dim(50,50), type, storage) 
   assert image.data.storage_format == storage
   assert image.data.pixel_type == type
   assert image.ul == Point(25, 25)
   assert image.size == Size(49, 49)
   #assert image.dimensions == Dimensions(50, 50) # deprecated call
   assert image.dim == Dim(50, 50)
   image2 = Image(Point(25, 25), Point(74, 74), type, storage)
   assert Rect(image) == Rect(image2)
   image3 = Image(Point(25, 25), Size(49, 49), type, storage)
   assert Rect(image) == Rect(image3)
   #image4 = Image(Point(25, 25), Dimensions(50, 50), type, storage) # deprecated call
   image4 = Image(Point(25, 25), Dim(50, 50), type, storage)
   assert Rect(image) == Rect(image4)
   #image5 = Image(Rect(25, 25, 50, 50), type, storage) # deprecated call
   image5 = Image(Rect(Point(25,25), Point(74,74)), type, storage)
   assert Rect(image) == Rect(image5)
   image6 = Image(Point(25, 25), Dim(50, 50), type, storage)
   assert Rect(image) == Rect(image6)
   py.test.raises(TypeError, _fail1)

test_image_constructors = make_test(_test_image_constructors)

def _test_subimage(type, value, storage):
   def _fail1():
      image.subimage(Point(20,20), Dim(10,10))
   def _fail2():
      image.subimage(Point(30,30), Dim(50,50))
   #image = Image(25, 25, 50, 50, type, storage) # deprecated call
   image = Image(Point(25,25), Dim(50,50), type, storage)
   #image2 = image.subimage(30, 30, 40, 40) # deprecated call
   image2 = image.subimage(Point(30,30), Dim(40,40))
   assert image2.ul == Point(30, 30)
   assert image2.data == image.data
   image3 = image.subimage(Point(30, 30), Point(69, 69))
   assert Rect(image2) == Rect(image3)
   assert image3.data == image.data
   image4 = image.subimage(Point(30, 30), Size(39, 39))
   assert Rect(image4) == Rect(image2)
   assert image4.data == image.data
   #image5 = image.subimage(Point(30, 30), Dimensions(40, 40)) # deprecated call
   image5 = image.subimage(Point(30, 30), Point(69, 69))
   assert Rect(image5) == Rect(image2)
   assert image5.data == image.data
   image6 = image.subimage(Rect(Point(30,30), Dim(40,40)))
   assert Rect(image6) == Rect(image2)
   assert image6.data == image.data
   image7 = image.subimage(Point(30, 30), Dim(40, 40))
   assert Rect(image7) == Rect(image2)
   assert image7.data == image.data
   py.test.raises(RuntimeError, _fail1)
   py.test.raises(RuntimeError, _fail2)
   #image7 = image2.subimage(35, 35, 20, 20) # deprecated call
   image7 = image2.subimage(Point(35,35), Dim(20,20))
   assert image7.data == image.data
   #image8 = image7.subimage(32, 32, 25, 25) # deprecated call
   image8 = image7.subimage(Point(32,32), Dim(25,25))
   assert image8.ul == Point(32, 32)
   assert image8.data == image.data
test_subimage = make_test(_test_subimage)

def _test_get_set(type, value, storage):
   def _fail1():
      image.set(Point(50, 50), 0)
   def _fail2():
      image.get(Point(50, 50))
   image = Image((25, 25), Dim(50, 50), type, storage)
   for val in value:
      image.set(Point(25, 25), val)
      assert image.get(Point(25, 25)) == val
   for r in range(50):
      for c in range(50):
         image.set((r, c), val)
         assert image.get((r, c)) == val
   py.test.raises(IndexError, _fail1)
   py.test.raises(IndexError, _fail2)
test_get_set = make_test(_test_get_set)
   
def _test_index(type, value, storage):
   def _fail1():
      image[50, 50] = 0
   def _fail2():
      image[50, 50]
   image = Image((25, 25), Dim(50, 50), type, storage)
   for val in value:
      image[25, 25] = val
      assert image[25, 25] == val
   for r in range(50):
      for c in range(50):
         image[r, c] = val
         assert image[r, c] == val
   print len(image)
   for i, p in enumerate(image):
      if p != val:
         print i
      assert p == val
   py.test.raises(IndexError, _fail1)
   py.test.raises(IndexError, _fail2)
test_index = make_test(_test_index)

def test_conversions():
    img = Image((0,0),(9,9),FLOAT)
    img.set((0,0),-5.0)
    img.set((9,9),10.0)
    # float -> grey
    tmp = img.to_greyscale()
    assert tmp.get((0,0)) == 0
    assert tmp.get((5,5)) == 85
    assert tmp.get((9,9)) == 255
    # float -> rgb
    tmp = img.to_rgb()
    assert tmp.get((0,0)) == RGBPixel(0,0,0)
    assert tmp.get((5,5)) == RGBPixel(85,85,85)
    assert tmp.get((9,9)) == RGBPixel(255,255,255)
    # float -> grey16
    tmp = img.to_grey16()
    assert tmp.get((0,0)) == 0
    assert tmp.get((5,5)) == 21845
    assert tmp.get((9,9)) == 65535
    # grey16 -> grey
    tmp = tmp.to_greyscale()
    assert tmp.get((0,0)) == 0
    assert tmp.get((5,5)) == 85
    assert tmp.get((9,9)) == 255



#
# Copyright (C) 2001-2007 Karl MacMillan, Michael Droettboom,
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Convert Images
#
# This script uses the wxPython img2py tool (included in this directory)
# to convert all of the pngs in the this directory into a python script
# called gamera_icons.py. This script can be copied to the main gamera
# directory to be used by gamera. This is easier to manage than the
# individual files, especially since distutils doesn't handle arbitrary
# data files.

import sys, glob, os
import img2py
import wx

if __name__ == "__main__":
    app = wx.App()

    files = glob.glob('*.png')
    first = 1
    for x in files:
        # This converts filenames in the form of file_name.png
        # into names of the form FileName suitable for method
        # names in the gamera_icons module
        base_name = x.split('.')[0]
        names = base_name.split('_')
        name = ''
        for j in names:
            name += j[0].upper() + j[1:]
        args = []
        # -a means append the image - only done after the first image
        if first == 0:
            args.append('-a')
        else:
            first = 0
        # -n sets the name pattern for the methods to access
        # this particular image
        args.append('-n')
        args.append(name)
        args.append(x)
        args.append('../gui/gamera_icons.py')
        # run the script
        img2py.main(args)



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os, os.path
from gamera import util
from gamera.core import *

class ClassifierStat:
   def __init__(self, classifier, path, max_size=64):
      self.classifier = classifier
      self.path = path
      if not os.path.exists(path):
         os.makedirs(path)
      self.image_path = os.path.join(path, "images")
      if not os.path.exists(self.image_path):
         os.makedirs(self.image_path)
      self.max_size = max_size

   def make_example_glyphs(self):
      self.example_glyphs = {}
      for glyph in self.classifier.get_glyphs():
         for conf, id in glyph.id_name:
            if not self.example_glyphs.has_key(id):
               self.example_glyphs[id] = glyph

   def make_grid(self, rows, cols):
      grid = []
      for i in range(rows):
         row = [None] * cols
         grid.append(row)
      return grid

   def make_pages(self):
      grids = self.make_result()
      for name, grid in grids:
         filename = os.path.join(self.path, name.lower().replace(" ", "_"))
         self.make_html(filename + ".html", name, grid)
         self.make_csv(filename + ".csv", name, grid)

   def make_html(self, filename, name, grid):
      fd = open(filename, "w")
      fd.write("<html><head><title>%s</title></head><body><h1>%s</h1>" %
               (name, name))
      fd.write("<table border=\"1\">")
      for row in grid:
         fd.write("<tr>")
         for col in row:
            fd.write("<td>")
            if isinstance(col, ImageBase):
               id = col.get_main_id()
               image_filename = "images/%s.png" % id
               col.save_PNG(os.path.join(self.path, image_filename))
               fd.write('<img src="%s" width="%d" height="%d"/><br/>%s' %
                        (image_filename, min(col.width, self.max_size),
                         min(col.height, self.max_size), id))
            elif col is None:
               fd.write("&nbsp;")
            else:
               fd.write(str(col))
            fd.write("</td>")
         fd.write("</tr>")
      fd.write("</table>")
      fd.write("</body></html>")
      fd.close()

   def make_csv(self, filename, name, grid):
      def convert(x):
         if isinstance(x, ImageBase):
            return x.get_main_id()
         elif x == None:
            return ""
         else:
            return str(x)
      fd = open(filename, "w")
      for row in grid:
         formatted_row = ", ".join([convert(x) for x in row])
         fd.write(formatted_row)
         fd.write("\n")
      fd.close()
            
class ConfusionMatrix(ClassifierStat):
   title = "Confusion Matrix"
   
   def make_result(self):
      self.make_example_glyphs()
      result = {}
      for id0 in self.example_glyphs.keys():
         leaf = {}
         for id1 in self.example_glyphs.keys():
            leaf[id1] = 0
         result[id0] = leaf
         
      classifier = self.classifier
      glyphs = classifier.get_glyphs()
      progress = util.ProgressFactory("Generating confusion matrix...", len(glyphs) / 50)
      try:
         for i, glyph in enumerate(glyphs):
            idname, conf = classifier.classify_with_images(glyphs, glyph, True)
            result[glyph.get_main_id()][idname[0][1]] += 1
            if i % 50 == 0:
               progress.step()
      finally:
         progress.kill()

      ids = result.keys()
      ids.sort()
      grid = self.make_grid(len(ids) + 1, len(ids) + 1)
      for i, id in enumerate(ids):
         grid[0][i+1] = self.example_glyphs[id]
         grid[i+1][0] = self.example_glyphs[id]
      for i, id0 in enumerate(ids):
         res = result[id0]
         sum = 0
         for val in res.values():
            sum += val
         for j, id0 in enumerate(ids):
            grid[i+1][j+1] = str(int((float(res[id0]) / sum) * 100.0)) + "%"
      return [("Confusion Matrix", grid)]

class ClassNameHistogram(ClassifierStat):
   title = "Class Name Histogram"

   def make_result(self):
      self.make_example_glyphs()
      result = {}
      for id0 in self.example_glyphs.keys():
         result[id0] = 0

      for glyph in self.classifier.get_glyphs():
         id = glyph.get_main_id()
         result[id] += 1

      result = [(val, key) for key, val in result.items()]
      result.sort()
      result.reverse()

      grid = self.make_grid(len(self.example_glyphs.keys()), 2)
      for i, (val, key) in enumerate(result):
         grid[i][0] = self.example_glyphs[key]
         grid[i][1] = val
      return [("Class Name Histogram", grid)]

all_stat_pages = [ConfusionMatrix, ClassNameHistogram]
def make_stat_pages(classifier, path, pages=None, max_size=64):
   if pages is None:
      pages = all_stat_pages
   for page in pages:
      name = page.__name__.lower()
      page_path = os.path.join(path, name)
      p = page(classifier, page_path, max_size)
      p.make_pages()




#
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2013      Christoph Dalitz, Fabian Schmitt
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import _runlength

# New version of functions.  Deprecated versions are below.

class most_frequent_run(PluginFunction):
    """
    Returns the length of the most frequently occurring run of pixels
    in the given color and given direction.
    """
    self_type = ImageType([ONEBIT])
    args = Args([ChoiceString("color", ["black", "white"]),
                 ChoiceString("direction", ["horizontal", "vertical"])])
    return_type = Int()
    doc_examples = [(ONEBIT, 'black', 'horizontal')]

class most_frequent_runs(PluginFunction):
    """
    Returns the lengths of the *n* most frequently occurring runs in
    the given *color* and *direction*.

    *n*
      The number of runlengths to return.  If *n* < 0, all runlengths
      will be returned.

    The return value is a list of 2-tuples.  The first element in the
    tuple is the run length, and the second element is its frequency.
    The list is sorted by descending frequency.
    """
    self_type = ImageType([ONEBIT])
    args = Args([Int("n"),
                 ChoiceString("color", ["black", "white"]),
                 ChoiceString("direction", ["horizontal", "vertical"])])
    return_type = Class()
    author = "Michael Droettboom, after an idea by Christoph Dalitz"
    def __call__(image, n = -1, color = 'black', direction = 'horizontal'):
        return _runlength.most_frequent_runs(image, n, color, direction)
    __call__ = staticmethod(__call__)
    doc_examples = [(ONEBIT, 5, 'black', 'horizontal')]

class run_histogram(PluginFunction):
    """
    Returns the histogram of runlengths in the given *color* and
    *direction*.

    *return_value*
      The return value is an integer array.  Each index in the array
      corresponds to a particular run length, and the value at that
      index is the number of times that that run length occurs in the
      image.
    """
    self_type = ImageType([ONEBIT])
    args = Args([ChoiceString("color", ["black", "white"]),
                 ChoiceString("direction", ["horizontal", "vertical"])])
    return_type = IntVector()
    doc_examples = [(ONEBIT, 'black', 'horizontal')]

class FilterRuns(PluginFunction):
    self_type = ImageType([ONEBIT])
    args = Args([Int("length"), ChoiceString("color", ["black", "white"])])
    doc_examples = [(ONEBIT, 3, 'black')]

class filter_narrow_runs(FilterRuns):
    """
    Removes horizontal runs in the given *color* narrower than a given
    *length*.
    """
    def __call__(image, length, color = 'black'):
        return _runlength.filter_narrow_runs(image, length, color)
    __call__ = staticmethod(__call__)

class filter_wide_runs(FilterRuns):
    """
    Removes horizontal runs in the given *color* wider than a given
    *length*.
    """
    def __call__(image, length, color = 'black'):
        return _runlength.filter_wide_runs(image, length, color)
    __call__ = staticmethod(__call__)

class filter_tall_runs(FilterRuns):
    """
    Removes vertical runs in the given *color* taller than a given
    *length*.
    """
    def __call__(image, length, color = 'black'):
        return _runlength.filter_tall_runs(image, length, color)
    __call__ = staticmethod(__call__)

class filter_short_runs(FilterRuns):
    """
    Removes vertical runs in the given *color* shorter than a given
    *length*.
    """
    def __call__(image, length, color = 'black'):
        return _runlength.filter_short_runs(image, length, color)
    __call__ = staticmethod(__call__)

class to_rle(PluginFunction):
    """
    Encodes a string-based run-length encoded version of the image.

    The numbers alternate between "length of black run" and "length of
    white run".  Runs go left-to-right, top-to-bottom.  Runs rollover
    the right hand edge and continue on the left edge of the next run.

    To decode an RLE string, use from_rle_.
    """
    self_type = ImageType([ONEBIT])
    return_type = String("runs")
    doc_examples = [(ONEBIT,)]

class from_rle(PluginFunction):
    """
    Decodes a string-based run-length encoded version of the image.

    The numbers alternate between "length of black run" and "length of
    white run".  Runs go left-to-right, top-to-bottom.  Runs rollover
    the right hand edge and continue on the left edge of the next run.

    To encode an RLE string, use to_rle_."""
    self_type = ImageType([ONEBIT])
    args = Args(String("runs"))

class iterate_runs(PluginFunction):
    """
    Returns nested iterators over the runs in the given *color* and
    *direction*.

    Each run is returned as a Rect object.
    
    For example, to iterate over all runs:

    .. code:: Python

      for row in image.iterate_black_horizontal_runs():
      # All the runs in each row
      for run in row:
          print run
    """
    self_type = ImageType([ONEBIT])
    args = Args([ChoiceString("color", ["black", "white"]),
                 ChoiceString("direction", ["horizontal", "vertical"])])
    return_type = Class()

class runlength_from_point(PluginFunction):
    """
    Returns the runlength of *color* in the direction *direction* starting
    from one pixel after *point*.
    """
    category = "Runlength"
    self_type = ImageType([ONEBIT])
    args = Args([FloatPoint("point"),ChoiceString("color", ["black", "white"]),
                 ChoiceString("direction", ["top", "bottom","left","right"])])
    return_type = Int()
    author = "Fabian Schmitt and Christoph Dalitz"


###########################################################################    
# Deprecated functions.

class FrequentRun(PluginFunction):
    self_type = ImageType([ONEBIT])
    return_type = Int()
    category = "Runlength/Deprecated"
    pure_python = True

class FrequentRuns(PluginFunction):
    self_type = ImageType([ONEBIT])
    args = Args([Int("n", default=-1)])
    return_type = Class()
    author = "Michael Droettboom, after an idea by Christoph Dalitz"
    category = "Runlength/Deprecated"
    pure_python = True

class RunHistogram(PluginFunction):
    self_type = ImageType([ONEBIT])
    return_type = IntVector()
    author = "Michael Droettboom"
    category = "Runlength/Deprecated"
    pure_python = True

class FilterRunsDep(PluginFunction):
    self_type = ImageType([ONEBIT])
    args = Args(Int("size"))
    category = "Runlength/Deprecated"
    pure_python = True

class RunIterator(PluginFunction):
    self_type = ImageType([ONEBIT])
    return_type = Class("iterator")
    category = "Runlength/Deprecated"
    pure_python = True

class RunLengthModule(PluginModule):
    category = "Runlength"
    cpp_headers=["runlength.hpp"]
    functions = [most_frequent_run,
                 most_frequent_runs,
                 run_histogram,
                 filter_narrow_runs,
                 filter_wide_runs,
                 filter_short_runs,
                 filter_tall_runs,
                 iterate_runs,
                 to_rle, from_rle,
                 runlength_from_point]

    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"

module = RunLengthModule()
                 
del FrequentRun
del FrequentRuns
del RunHistogram
del FilterRuns
del RunIterator
del FilterRunsDep



#
# This file has been taken from wxpython (see the file
# wx/tools/img2py.py in the wxpython source distribution)
#
# Copyright (c) 1998 Julian Smart, Robert Roebling et al
#
# This program may be freely used, copied and distributed under
# the terms of the wxWindows Library Licence, Version 3. See
# the file "copyright" of the wxpython distribution from
# http://wxpython.org/ for details.
#
"""
img2py.py  --  Convert an image to PNG format and embed it in a Python
               module with appropriate code so it can be loaded into
               a program at runtime.  The benefit is that since it is
               Python source code it can be delivered as a .pyc or
               'compiled' into the program using freeze, py2exe, etc.

Usage:

    img2py.py [options] image_file python_file

Options:

    -m <#rrggbb>   If the original image has a mask or transparency defined
                   it will be used by default.  You can use this option to
                   override the default or provide a new mask by specifying
                   a colour in the image to mark as transparent.

    -n <name>      Normally generic names (getBitmap, etc.) are used for the
                   image access functions.  If you use this option you can
                   specify a name that should be used to customize the access
                   fucntions, (getNameBitmap, etc.)

    -a             This flag specifies that the python_file should be appended
                   to instead of overwritten.  This in combination with -n will
                   allow you to put multiple images in one Python source file.

    -u             Don't use compression.  Leaves the data uncompressed.

    -i             Also output a function to return the image as a wxIcon.

"""



import sys, os, glob, getopt, tempfile, string
import cPickle, cStringIO, zlib
import img2img
try:
    import wx
except Exception:
    from wxPython import wx


def crunch_data(data, compressed):
    # compress it?
    if compressed:
        data = zlib.compress(data, 9)

    # convert to a printable format, so it can be in a Python source file
    data = repr(data)

    # This next bit is borrowed from PIL.  It is used to wrap the text intelligently.
    fp = cStringIO.StringIO()
    data = data + " "  # buffer for the +1 test
    c = i = 0
    word = ""
    octdigits = "01234567"
    hexdigits = "0123456789abcdef"
    while i < len(data):
        if data[i] != "\\":
            word = data[i]
            i = i + 1
        else:
            if data[i+1] in octdigits:
                for n in range(2, 5):
                    if data[i+n] not in octdigits:
                        break
                word = data[i:i+n]
                i = i + n
            elif data[i+1] == 'x':
                for n in range(2, 5):
                    if data[i+n] not in hexdigits:
                        break
                word = data[i:i+n]
                i = i + n
            else:
                word = data[i:i+2]
                i = i + 2

        l = len(word)
        if c + l >= 78-1:
            fp.write("\\\n")
            c = 0
        fp.write(word)
        c = c + l

    # return the formatted compressed data
    return fp.getvalue()



def main(args):
    if not args or ("-h" in args):
        print __doc__
        return

    append = 0
    compressed = 1
    maskClr = None
    imgName = ""
    icon = 0

    try:
        opts, fileArgs = getopt.getopt(args, "auin:m:")
    except getopt.GetoptError:
        print __doc__
        return

    for opt, val in opts:
        if opt == "-a":
            append = 1
        elif opt == "-u":
            compressed = 0
        elif opt == "-n":
            imgName = val
        elif opt == "-m":
            maskClr = val
        elif opt == "-i":
            icon = 1

    if len(fileArgs) != 2:
        print __doc__
        return

    image_file, python_file = fileArgs

    # convert the image file to a temporary file
    tfname = tempfile.mktemp()
    ok, msg = img2img.convert(image_file, maskClr, None, tfname, wx.BITMAP_TYPE_PNG, ".png")
    if not ok:
        print msg
        return

    data = open(tfname, "rb").read()
    data = crunch_data(data, compressed)
    os.unlink(tfname)

    if append:
        out = open(python_file, "a")
    else:
        out = open(python_file, "w")

    out.write("#" + "-" * 70 + "\n")
    if not append:
        out.write("# This file was generated by %s\n#\n" % sys.argv[0])
        out.write("try:\n")
        out.write("   from wx import ImageFromStream, BitmapFromImage\n")
        out.write("   wxImageFromStream = ImageFromStream\n")
        out.write("   wxBitmapFromImage = BitmapFromImage\n")
        out.write("except:\n")
        out.write("   from wxPython.wx import wxImageFromStream, wxBitmapFromImage\n")
        if compressed:
            out.write("import cStringIO, zlib\n\n\n")
        else:
            out.write("import cStringIO\n\n\n")

    if compressed:
        out.write("def get%sData():\n"
                  "    return zlib.decompress(\n%s)\n\n"
                  % (imgName, data))
    else:
        out.write("def get%sData():\n"
                  "    return \\\n%s\n\n"
                  % (imgName, data))


    out.write("def get%sBitmap():\n"
              "    return wxBitmapFromImage(get%sImage())\n\n"
              "def get%sImage():\n"
              "    stream = cStringIO.StringIO(get%sData())\n"
              "    return wxImageFromStream(stream)\n\n"
              % tuple([imgName] * 4))
    if icon:
        out.write("def get%sIcon():\n"
                  "    icon = wxEmptyIcon()\n"
                  "    icon.CopyFromBitmap(get%sBitmap())\n"
                  "    return icon\n\n"
                  % tuple([imgName] * 2))


    if imgName:
        n_msg = ' using "%s"' % imgName
    else:
        n_msg = ""
    if maskClr:
        m_msg = " with mask %s" % maskClr
    else:
        m_msg = ""
    print "Embedded %s%s into %s%s" % (image_file, n_msg, python_file, m_msg)

if __name__ == "__main__":
    main(sys.argv[1:])




# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.gui import has_gui

if has_gui.has_gui:
   from gamera.gui import gui
   CustomMenu = gui.CustomMenu
   from gamera.gui import icon_display
   CustomIcon = icon_display.CustomIcon
else:
   class NullClass:
      def __init__(self, *args, **kwargs):
         pass
   class CustomMenu(NullClass):
      pass
   class CustomIcon(NullClass):
      def register(cls, *args, **kwargs):
         pass
      register = classmethod(register)

__all__ = "CustomMenu CustomIcon".split()



#
# Copyright (C) 2012 Tobias Bolten
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx

from gamera import knnga

#-------------------------------------------------------------------------------
class ResultPanel(wx.ScrolledWindow):
#-------------------------------------------------------------------------------
    #---------------------------------------------------------------------------
    def __init__(self, parent, id, frame):
    #---------------------------------------------------------------------------
        wx.ScrolledWindow.__init__(self, parent, -1)

        self.frame = frame
        self.bars = None

    #---------------------------------------------------------------------------
    def initResultsWidgets(self):
    #---------------------------------------------------------------------------
        if self.frame.classifier == None:
            raise ValueError("ResultPanel.initResultsWidgets: no valid classifier given")

        for child in self.GetChildren():
            child.Destroy()

        features = self.frame.classifier.feature_functions
        feature_names = [x[0] for x in features[0]]
        feature_functions = [x[1] for x in features[0]]

        if self.frame.settingsPanel.featureSelection.GetValue():
            values = self.frame.classifier.get_selections_by_features()
        elif self.frame.settingsPanel.featureWeighting.GetValue():
            values = self.frame.classifier.get_weights_by_features()
        else:
            raise ValueError("ResultPanel.initResultsWidgets: unknown mode of operation")

        gridSizer = wx.FlexGridSizer(0, 2, 5, 15)
        gridSizer.AddGrowableCol(1)

        flags = (wx.TOP|wx.LEFT|wx.RIGHT|wx.EXPAND|wx.ALIGN_CENTER_VERTICAL)

        text = wx.StaticText(self, -1, "Feature Name")
        gridSizer.Add(text, 0, flags, 10)
        text = wx.StaticText(self, -1, "Value")
        gridSizer.Add(text, 0, flags, 10)

        for (name, ff) in zip(feature_names, feature_functions):
            self.bars[name] = []
            for x in range(ff.return_type.length):

                feature_name = wx.StaticText(self, -1, name + " - " + str(x))
                gridSizer.Add(feature_name, 0, flags, 10)

                bar = wx.Gauge(self, -1, 100, style=wx.GA_SMOOTH, size=(-1,10))
                bar.SetValue(values[name][x]*100)
                gridSizer.Add(bar, 0, flags, 10)

                self.bars[name].append(bar)

        self.SetSizer(gridSizer)

        width, height = gridSizer.GetMinSize()
        self.SetScrollbars(10, 10, width/10, height/10)

        self.Layout()

    #---------------------------------------------------------------------------
    def Update(self):
    #---------------------------------------------------------------------------
        wx.CallAfter(self.UpdateResultWidgets)

    #---------------------------------------------------------------------------
    def UpdateResultWidgets(self):
    #---------------------------------------------------------------------------
        if self.frame.classifier == None:
            return

        if self.bars == None:
            self.bars = {}
            self.initResultsWidgets()

        features = self.frame.classifier.feature_functions
        feature_names = [x[0] for x in features[0]]
        feature_functions = [x[1] for x in features[0]]

        if self.frame.settingsPanel.featureSelection.GetValue():
            values = self.frame.classifier.get_selections_by_features()
        elif self.frame.settingsPanel.featureWeighting.GetValue():
            values = self.frame.classifier.get_weights_by_features()
        else:
            raise ValueError("ResultPanel.UpdateResultWidgets: unknown mode of operation")

        if len([x for x in self.bars.keys() if x not in feature_names]) != 0:
            self.initResultsWidgets()
            return

        if len([x for x in feature_names if x not in self.bars.keys()]) != 0:
            self.initResultsWidgets()
            return

        for (name, ff) in zip(feature_names, feature_functions):
            for x in range(ff.return_type.length):
                self.bars[name][x].SetValue(values[name][x]*100)



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:

"""PyPlate : a simple Python-based templating program

PyPlate parses a file and replaces directives (in double square brackets [[ ... ]])
by various means using a given dictionary of variables.  Arbitrary Python code
can be run inside many of the directives, making this system highly flexible.

Usage:
# Load and parse template file
template = pyplate.Template("output") (filename or string)
# Execute it with a dictionary of variables
template.execute_file(output_stream, locals())

PyPlate defines the following directives:
   [[...]]       evaluate the arbitrary Python expression and insert the
      result into the output

   [[# ... #]]   comment.

   [[exec ...]]  execute arbitrary Python code in the sandbox namespace

   [[if ...]]    conditional expressions with usual Python semantics
   [[elif ...]]
   [[else]]
   [[end]]

   [[for ... in ...]]  for-loop with usual Python semantics
   [[end]]

   [[def ...(...)]]  define a "function" out of other templating elements
   [[end]]

   [[call ...]]  call a templating function (not a regular Python function)
 """

#
# Copyright (C) 2001-2005 Michael Droettboom
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import nested_scopes
import sys, string, re, util, cStringIO, codecs

re_directive = re.compile("\[\[(.*?)\]\]")
re_for_loop = re.compile("for (.*) in (.*)")
re_if = re.compile("if (.*)")
re_elif = re.compile("elif (.*)")
re_def = re.compile("def (.*?)\((.*)\)")
re_call = re.compile("call (.*?)\((.*)\)")
re_exec = re.compile("exec (.*)")
re_comment = re.compile("#(.*)#")

re_clean_whitespace = re.compile(r"\]\]\s+?")

############################################################
# Template parser
class ParserException(Exception):
   def __init__(self, lineno, s):
      Exception.__init__(self, "line %d: %s" % (lineno, s))

class Template:
   def __init__(self, filename=None):
      if filename != None:
         try:
            self.parse_file(filename)
         except Exception:
            self.parse_string(filename)

   def parse_file(self, filename):
      file = open(filename, 'r')
      self.parse(file)
      file.close()

   def parse_string(self, template):
      file = cStringIO.StringIO(template)
      self.parse(file)
      file.close()

   def parse(self, file):
      self.file = file
      self.line = re_clean_whitespace.sub("]]", self.file.read())
      self.lineno = 0
      self.functions = {}
      self.tree = TopLevelTemplateNode(self)

   def parser_get(self):
      if self.line == '':
         return None
      return self.line

   def parser_eat(self, chars):
      self.lineno = self.lineno + self.line[:chars].count("\n")
      self.line = self.line[chars:]

   def parser_exception(self, s, e=None):
      import traceback
      if e != None:
         traceback.print_exc()
      raise ParserException(self.lineno, s)

   def execute_file(self, filename, data={}):
      output = self.execute_string(data)
      file = open(filename, 'w')
      file.write(output)
      file.write("\n")
      file.close()

   def execute_string(self, data={}):
      data_copy = {}
      data_copy.update(data)
      s = cStringIO.StringIO()
      self.execute(s, data_copy)
      s.write("\n")
      return s.getvalue()

   def execute_stdout(self, data={}):
      data_copy = {}
      data_copy.update(data)
      self.execute(sys.stdout, data_copy)
      sys.stdout.write("\n")

   def execute(self, stream=sys.stdout, data={}):
      self.tree.execute(stream, data)

   def __repr__(self):
      return repr(self.tree)


############################################################
# NODES
class TemplateNode:
   def __init__(self, parent, s):
      self.parent = parent
      self.s = s
      self.node_list = []
      while 1:
         new_node = TemplateNodeFactory(parent)
         if self.add_node(new_node):
            break

   def add_node(self, node):
      if node == 'end':
         return 1
      elif node != None:
         self.node_list.append(node)
      else:
         raise self.parent.parser_exception(
           "[[%s]] does not have a matching [[end]]" % self.s)

   def execute(self, stream, data):
      for node in self.node_list:
         node.execute(stream, data)

   def __repr__(self):
      r = "<" + self.__class__.__name__ + " "
      for i in self.node_list:
         r = r + repr(i)
      r = r + ">"
      return r

class TopLevelTemplateNode(TemplateNode):
   def __init__(self, parent):
      TemplateNode.__init__(self, parent, '')

   def add_node(self, node):
      if node != None:
         self.node_list.append(node)
      else:
         return 1

class ForTemplateNode(TemplateNode):
   def __init__(self, parent, s):
      TemplateNode.__init__(self, parent, s)
      match = re_for_loop.match(s)
      if match == None:
         raise self.parent.parser_exception(
           "[[%s]] is not a valid for-loop expression" % self.s)
      else:
         self.vars_temp = match.group(1).split(",")
         self.vars = []
         for v in self.vars_temp:
            self.vars.append(v.strip())
         self.expression = match.group(2)

   def execute(self, stream, data):
      remember_vars = {}
      for var in self.vars:
         if data.has_key(var):
            remember_vars[var] = data[var]
      try:
         x = eval(self.expression, globals(), data)
      except Exception, e:
         self.parent.parser_exception(self.expression, e)
      for list in x:
         if util.is_sequence(list):
            if len(self.vars) == 1:
               data[self.vars[0]] = list
            elif len(self.vars) == len(list):
               for index, value in util.enumerate(list):
                  data[self.vars[index]] = value
            else:
               self.parent.parser_exception(
                 "Unable to unpack tuples in [[%s]]" % self.s)
         else:
            data[self.vars[0]] = list
         TemplateNode.execute(self, stream, data)
      for key, value in remember_vars.items():
         data[key] = value

class IfTemplateNode(TemplateNode):
   def __init__(self, parent, s):
      self.else_node = None
      TemplateNode.__init__(self, parent, s)
      match = re_if.match(s)
      if match == None:
         raise self.parent.parser_exception(
           "[[%s]] is not a valid if expression" % self.s)
      else:
         self.expression = match.group(1)

   def add_node(self, node):
      if node == 'end':
         return 1
      elif isinstance(node, ElseTemplateNode):
         self.else_node = node
         return 1
      elif isinstance(node, ElifTemplateNode):
         self.else_node = node
         return 1
      elif node != None:
         self.node_list.append(node)
      else:
         raise self.parent.parser_exception(
           "[[%s]] does not have a matching [[end]]" % self.s)

   def execute(self, stream, data):
      try:
         x = eval(self.expression, globals(), data)
      except Exception, e:
         self.parent.parser_exception(self.expression, e)
      if x:
         TemplateNode.execute(self, stream, data)
      elif self.else_node != None:
         self.else_node.execute(stream, data)

class ElifTemplateNode(IfTemplateNode):
   def __init__(self, parent, s):
      self.else_node = None
      TemplateNode.__init__(self, parent, s)
      match = re_elif.match(s)
      if match == None:
         self.parent.parser_exception(
           "[[%s]] is not a valid elif expression" % self.s)
      else:
         self.expression = match.group(1)

class ElseTemplateNode(TemplateNode):
   pass

class FunctionTemplateNode(TemplateNode):
   def __init__(self, parent, s):
      TemplateNode.__init__(self, parent, s)
      match = re_def.match(s)
      if match == None:
         self.parent.parser_exception(
           "[[%s]] is not a valid function definition" % self.s)
      self.function_name = match.group(1)
      self.vars_temp = match.group(2).split(",")
      self.vars = []
      for v in self.vars_temp:
         self.vars.append(v.strip())
      self.parent.functions[self.function_name] = self

   def execute(self, stream, data):
      pass

   def call(self, args, stream, data):
      remember_vars = {}
      for index, var in util.enumerate(self.vars):
         if data.has_key(var):
            remember_vars[var] = data[var]
         data[var] = args[index]
      TemplateNode.execute(self, stream, data)
      for key, value in remember_vars.items():
         data[key] = value

class LeafTemplateNode(TemplateNode):
   def __init__(self, parent, s):
      self.parent = parent
      self.s = s

   def execute(self, stream, data):
      stream.write(self.s)

   def __repr__(self):
      return "<" + self.__class__.__name__ + ">"

class CommentTemplateNode(LeafTemplateNode):
   def execute(self, stream, data):
      pass

class ExpressionTemplateNode(LeafTemplateNode):
   def execute(self, stream, data):
      try:
         stream.write(unicode(eval(self.s, globals(), data)).encode("utf-8"))
      except Exception, e:
         self.parent.parser_exception(self.s, e)

class ExecTemplateNode(LeafTemplateNode):
   def __init__(self, parent, s):
      LeafTemplateNode.__init__(self, parent, s)
      match = re_exec.match(s)
      if match == None:
         self.parent.parser_exception(
           "[[%s]] is not a valid statement" % self.s)
      self.s = match.group(1)

   def execute(self, stream, data):
      try:
         exec(self.s, globals(), data)
      except Exception, e:
         self.parent.parser_exception(self.s, e)

class CallTemplateNode(LeafTemplateNode):
   def __init__(self, parent, s):
      LeafTemplateNode.__init__(self, parent, s)
      match = re_call.match(s)
      if match == None:
         self.parent.parser_exception(
           "[[%s]] is not a valid function call" % self.s)
      self.function_name = match.group(1)
      self.vars = "(" + match.group(2).strip() + ",)"

   def execute(self, stream, data):
      try:
         x = eval(self.vars, globals(), data)
      except Exception, e:
         self.parent.parser_exception(self.vars, e)
      self.parent.functions[self.function_name].call(
        x, stream, data)


############################################################
# Node factory
template_factory_type_map = {
  'if'   : IfTemplateNode,
  'for'  : ForTemplateNode,
  'elif' : ElifTemplateNode,
  'else' : ElseTemplateNode,
  'def'  : FunctionTemplateNode,
  'call' : CallTemplateNode,
  'exec' : ExecTemplateNode }
template_factory_types = template_factory_type_map.keys()

def TemplateNodeFactory(parent):
   src = parent.parser_get()

   if src == None:
      return None
   match = re_directive.search(src)
   if match == None:
      parent.parser_eat(len(src))
      return LeafTemplateNode(parent, src)
   elif src == '' or match.start() != 0:
      parent.parser_eat(match.start())
      return LeafTemplateNode(parent, src[:match.start()])
   else:
      directive = match.group()[2:-2].strip()
      parent.parser_eat(match.end())
      if directive == 'end':
         return 'end'
      elif re_comment.match(directive):
         return CommentTemplateNode(parent, directive)
      else:
         for i in template_factory_types:
            if directive[0:len(i)] == i:
               return template_factory_type_map[i](parent, directive)
         return ExpressionTemplateNode(parent, directive)


############################################################
# TESTING CODE
if __name__ == '__main__':
   combinations = (('OneBit', 'Float', 'GreyScale'),
                   ('GreyScale', 'RGB'))

   template = Template("""
   [[# This is a comment #]]
   [[# This example does recursive function calls need to generate feature combinations #]]
   [[def switch(layer, args)]]
      switch(m[[layer]].id) {
      [[for option in combinations[layer] ]]
      [[exec current = option + '(m' + str(layer) + ')']]
      case [[option]]:
        [[if layer == layers - 1]]
          function_call([[string.join(args + [current], ',')]]);
        [[else]]
          [[call switch(layer + 1, args + [current])]]
        [[end]]
      break;
      [[end]]
      }
   [[end]]

   PyObject *py_overload_resolution_[[function_name]](PyObject *args) {
   [[call switch(0, [])]]
   }
   """)

   data = {'combinations'  : combinations,
           'function_name' : 'threshold',
           'layers'        : 2}
   template.execute(sys.stdout, data)
